"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import abc
import asyncio
import functools
import logging
import pickle
import struct
import traceback
import typing
import uuid
from asyncio import Future, Protocol, transports
from dataclasses import dataclass
from pickle import PicklingError
from typing import Optional


class IPCException(Exception):
    pass


class ConnectionLost(IPCException):
    pass


ServerContext = typing.TypeVar("ServerContext")
""" An object to give IPC calls access to the server """

ReturnType = typing.TypeVar("ReturnType")


class IPCFrame(abc.ABC):
    """Interface marker for IPC frames"""

    pass


class IPCMethod(IPCFrame, typing.Generic[ServerContext, ReturnType]):
    """Base class for methods intended for IPC"""

    @abc.abstractmethod
    async def call(self, context: ServerContext) -> ReturnType:
        pass


@dataclass
class IPCRequestFrame(IPCFrame, typing.Generic[ServerContext, ReturnType]):
    id: Optional[uuid.UUID]
    method: IPCMethod[ServerContext, ReturnType]


@dataclass
class IPCReplyFrame(IPCFrame):
    id: uuid.UUID
    returnvalue: object
    is_exception: bool


@dataclass
class IPCLogRecord(IPCFrame):
    """
    Derived from logging.LogRecord, but simplified

    :param name: the logger name, as on logging.LogRecord
    :param levelno: the log level, in numeric form, as on logging.LogRecord
    :param msg: the message, as produced by record.getMessage() i.e. this record has all arguments already formatted in
    """

    name: str
    levelno: int
    msg: str


class IPCFrameProtocol(Protocol):
    """
    Simple protocol which sends

    frame_length: 4 bytes, unsigned integer
    frame: frame_length, pickled data

    It is intended as an async replacement of the facilities offered by multiprocessing and of similar design.

    This protocol is only suited for local interprocess communications, when both ends are trusted.
    This protocol is based on pickle for speed, so it is as insecure as pickle.
    """

    def __init__(self, name: str) -> None:
        # Expected size of frame
        # -1 if no frame in flight
        self.frame_size = -1

        # Buffer with all data we have received and not dispatched
        # We could sqeeze out some more performance by using memoryview instead of an array
        self.frame_buffer: Optional[bytes] = None

        # Our transport
        self.transport: Optional[transports.Transport] = None

        self.name = name
        self.logger = logging.getLogger(f"ipc.{name}")

    def connection_made(self, transport: transports.Transport) -> None:
        # Capture the transport
        self.transport = transport

    def data_received(self, data: bytes) -> None:
        # Get a block of data
        # Append to frame buffer
        if self.frame_buffer is not None:
            self.frame_buffer += data
        else:
            self.frame_buffer = data
        while True:
            # Eat up frames
            if self.frame_size == -1:
                if len(self.frame_buffer) < 4:
                    # incomplete length field, wait for data
                    break

                # new frame length received
                length = struct.unpack_from("!L", self.frame_buffer)[0]
                self.frame_size = length
            if len(self.frame_buffer) >= self.frame_size + 4:
                # Fill frame in buffer, dispatch
                self._block_received(self.frame_buffer[4 : self.frame_size + 4])
                # Truncate buffer
                self.frame_buffer = self.frame_buffer[self.frame_size + 4 :]
                # Reset frame size
                self.frame_size = -1
            else:
                # Not full frame anymore, wait for data
                break

    def _block_received(self, block: bytes) -> None:
        """Interception point for tests of block handling"""
        try:
            frame = pickle.loads(block)
        except Exception:
            # Failed to unpickle, drop connection
            self.logger.exception("Dropping IPC connection %s because of deserialization failure", self.name)
            self.transport.close()
            return
        try:
            self.frame_received(frame)
        except Exception:
            # Failed to unpickle, drop frame
            self.logger.exception("Unexpected exception while handling frame %s", self.name)

    def send_frame(self, frame: IPCFrame) -> None:
        """
        Helper method to construct and send frames
        """
        if self.transport.is_closing():
            raise ConnectionLost()
        try:
            buffer = pickle.dumps(frame)
        except PicklingError:
            raise
        except Exception as e:
            # Pickle tends to raise other exceptions as well...
            raise PicklingError() from e
        size = struct.pack("!L", len(buffer))
        self.transport.write(size + buffer)

    def frame_received(self, frame: IPCFrame) -> None:
        """
        Method for frame handling subclasses

        Always call super, to use multiple inheritance to compose handlers
        """
        raise Exception(f"Frame not handled {frame}")


class IPCServer(IPCFrameProtocol, abc.ABC, typing.Generic[ServerContext]):
    """Base server that dispatched methods"""

    @abc.abstractmethod
    def get_context(self) -> ServerContext:
        pass

    # TODO: timeouts?
    def frame_received(self, frame: IPCFrame) -> None:
        if isinstance(frame, IPCRequestFrame):
            asyncio.get_running_loop().create_task(self.dispatch(frame))
        else:
            super().frame_received(frame)

    async def dispatch(self, frame: IPCRequestFrame[ServerContext, ReturnType]) -> None:
        """
        Dispatch handler that sends back return values
        """
        try:
            return_value = await frame.method.call(self.get_context())
            if frame.id is not None:
                self.send_frame(IPCReplyFrame(frame.id, return_value, is_exception=False))
        except ConnectionLost:
            self.logger.debug("Connection lost", exc_info=True)
        except Exception as e:
            self.logger.debug("Exception on rpc call", exc_info=True)
            if frame.id is not None:
                try:
                    self.send_frame(IPCReplyFrame(frame.id, e, is_exception=True))
                except PicklingError:
                    # Can't pickle it

                    self.send_frame(IPCReplyFrame(frame.id, traceback.format_exception(e), is_exception=True))


class IPCClient(IPCFrameProtocol, typing.Generic[ServerContext]):
    """Base client that dispatched method calls"""

    def __init__(self, name: str):
        super().__init__(name)
        # TODO timeouts

        # All outstanding calls
        self.requests: dict[uuid.UUID, Future[object]] = {}

    @typing.overload
    def call(
        self, method: IPCMethod[ServerContext, ReturnType], has_reply: typing.Literal[True] = True
    ) -> Future[ReturnType]: ...

    @typing.overload
    def call(self, method: IPCMethod[ServerContext, ReturnType], has_reply: typing.Literal[False]) -> None: ...

    @typing.overload
    def call(self, method: IPCMethod[ServerContext, ReturnType], has_reply: bool = True) -> Future[ReturnType] | None: ...

    def call(self, method: IPCMethod[ServerContext, ReturnType], has_reply: bool = True) -> Future[ReturnType] | None:
        """Call a method with given arguments"""
        request = IPCRequestFrame(
            id=uuid.uuid4() if has_reply else None,
            method=method,
        )
        self.send_frame(request)

        if not has_reply:
            return None

        done = asyncio.get_event_loop().create_future()
        self.requests[request.id] = done  # Mypy can't do it
        return done

    def frame_received(self, frame: IPCFrame) -> None:
        """Handle replies"""
        if isinstance(frame, IPCReplyFrame):
            self.process_reply(frame)
        else:
            super().frame_received(frame)

    def process_reply(self, frame: IPCReplyFrame) -> None:
        if frame.is_exception:
            if isinstance(frame.returnvalue, Exception):
                self.requests[frame.id].set_exception(frame.returnvalue)
            else:
                self.requests[frame.id].set_exception(Exception(frame.returnvalue))
        else:
            self.requests[frame.id].set_result(frame.returnvalue)
        del self.requests[frame.id]

    def connection_lost(self, exc: Exception | None) -> None:
        excn = ConnectionLost()
        excn.__cause__ = exc
        for outstanding_request in self.requests.values():
            outstanding_request.set_exception(excn)
        self.requests.clear()
        super().connection_lost(exc)


class FinalizingIPCClient(IPCClient[ServerContext]):
    """IPC client that can signal shutdown"""

    def __init__(self, name: str):
        super().__init__(name)
        self.finalizers: list[typing.Callable[[], typing.Coroutine[typing.Any, typing.Any, None]]] = []
        # Collection to avoid task getting garbage collected
        self.finalizer_anti_gc: set[asyncio.Task[None]] = set()

    def connection_lost(self, exc: Exception | None) -> None:
        super().connection_lost(exc)
        for fin in self.finalizers:
            task = asyncio.get_running_loop().create_task(fin())
            self.finalizer_anti_gc.add(task)
            task.add_done_callback(self.finalizer_anti_gc.discard)


class LogReceiver(IPCFrameProtocol):
    """
    IPC feature to receive log message

    It re-injects the log message into the logging framework in the exact same place as it was on the sender side.

    This makes the LogShipper/LogReceiver pair a (mostly) transparent bridge.
    Log records are simplified when transported.

    When installing the LogShipper and LogReceiver in the same process, this will create an infinite loop
    """

    def frame_received(self, frame: IPCFrame) -> None:
        if isinstance(frame, IPCLogRecord):
            # calling log here is safe because if there are no arguments, formatter is never called
            logging.getLogger(frame.name).log(frame.levelno, frame.msg)
        else:
            super().frame_received(frame)


class LogShipper(logging.Handler):
    """
    Log sender associated with the log receiver

    This sender is threadsafe
    """

    def __init__(self, protocol: IPCFrameProtocol, eventloop: asyncio.AbstractEventLoop) -> None:
        self.protocol = protocol
        self.eventloop = eventloop
        self.logger_name = "inmanta.ipc.logs"
        self.logger = logging.getLogger(self.logger_name)
        super().__init__()

    def _send_frame(self, record: IPCLogRecord) -> None:
        try:
            self.protocol.send_frame(record)
        except ConnectionLost:
            # Stop exception here
            # Log in own logger to prevent loops
            self.logger.debug("Could not send log line, connection lost %s", record.msg, exc_info=True)
            return
        except Exception:
            # Stop exception here
            # Log in own logger to prevent loops
            self.logger.info("Could not send log line %s", record.msg, exc_info=True)
            return

    def emit(self, record: logging.LogRecord) -> None:
        if record.name == self.logger_name:
            # avoid loops
            # When we fail to send, we produce a log line on this logger
            return
        self.eventloop.call_soon_threadsafe(
            functools.partial(
                self._send_frame,
                IPCLogRecord(
                    record.name,
                    record.levelno,
                    self.format(record),
                ),
            )
        )



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import os
import signal
import sys
import threading
import traceback
from threading import Timer
from types import FrameType
from typing import Any, Callable, Coroutine, Optional

from tornado import gen
from tornado.ioloop import IOLoop
from tornado.util import TimeoutError

from inmanta import const

try:
    import rpdb
except ImportError:
    rpdb = None


def dump_threads() -> None:
    print("----- Thread Dump ----")
    for th in threading.enumerate():
        print("---", th)
        if th.ident:
            traceback.print_stack(sys._current_frames()[th.ident], file=sys.stdout)
        print()
    sys.stdout.flush()


async def dump_ioloop_running() -> None:
    # dump async IO
    print("----- Async IO tasks ----")
    for task in asyncio.all_tasks():
        print(task)
    print()
    sys.stdout.flush()


def context_dump(ioloop: IOLoop) -> None:
    dump_threads()
    if hasattr(asyncio, "all_tasks"):
        ioloop.add_callback_from_signal(dump_ioloop_running)


def setup_signal_handlers(shutdown_function: Callable[[], Coroutine[Any, Any, None]]) -> None:
    """
    Make sure that shutdown_function is called when a SIGTERM or a SIGINT interrupt occurs.

    :param shutdown_function: The function that contains the shutdown logic.
    """
    # ensure correct ioloop
    ioloop = IOLoop.current()

    def hard_exit() -> None:
        context_dump(ioloop)
        sys.stdout.flush()
        # Hard exit, not sys.exit
        # ensure shutdown when the ioloop is stuck
        os._exit(const.EXIT_HARD)

    def handle_signal(signum: signal.Signals, frame: Optional[FrameType]) -> None:
        # force shutdown, even when the ioloop is stuck
        # schedule off the loop
        t = Timer(const.SHUTDOWN_GRACE_HARD, hard_exit)
        t.daemon = True
        t.start()
        ioloop.add_callback_from_signal(safe_shutdown_wrapper, shutdown_function)

    def handle_signal_dump(signum: signal.Signals, frame: Optional[FrameType]) -> None:
        context_dump(ioloop)

    signal.signal(signal.SIGTERM, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGUSR1, handle_signal_dump)
    if rpdb:
        rpdb.handle_trap()


def safe_shutdown(ioloop: IOLoop, shutdown_function: Callable[[], None]) -> None:
    def hard_exit() -> None:
        context_dump(ioloop)
        sys.stdout.flush()
        # Hard exit, not sys.exit
        # ensure shutdown when the ioloop is stuck
        os._exit(const.EXIT_HARD)

    # force shutdown, even when the ioloop is stuck
    # schedule off the loop
    t = Timer(const.SHUTDOWN_GRACE_HARD, hard_exit)
    t.daemon = True
    t.start()
    ioloop.add_callback(safe_shutdown_wrapper, shutdown_function)


async def safe_shutdown_wrapper(shutdown_function: Callable[[], Coroutine[Any, Any, None]]) -> None:
    """
    Wait 10 seconds to gracefully shutdown the instance.
    Afterwards stop the IOLoop
    Wait for 3 seconds to force stop
    """
    future = shutdown_function()
    try:
        timeout = IOLoop.current().time() + const.SHUTDOWN_GRACE_IOLOOP
        await gen.with_timeout(timeout, future)
    except TimeoutError:
        pass
    finally:
        IOLoop.current().stop()



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from enum import Enum

from inmanta.data import model


class EnvironmentListener:
    """
    Base class for environment listeners
    Exceptions from the listeners are dropped, the listeners are responsible for handling them
    """

    async def environment_action_created(self, env: model.Environment) -> None:
        """
        Will be called when a new environment is created
        :param env: The new environment
        """

    async def environment_action_cleared(self, env: model.Environment) -> None:
        """
        Will be called when the environment is cleared
        :param env: The environment that is cleared
        """

    async def environment_action_deleted(self, env: model.Environment) -> None:
        """
        Will be called when the environment is deleted
        :param env: The environment that is deleted
        """

    async def environment_action_updated(self, updated_env: model.Environment, original_env: model.Environment) -> None:
        """
        Will be called when an environment is updated
        :param updated_env: The updated environment
        :param original_env: The original environment
        """


class EnvironmentAction(str, Enum):
    created = "created"
    deleted = "deleted"
    cleared = "cleared"
    updated = "updated"



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import traceback
from abc import abstractmethod
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from inmanta import warnings
from inmanta.ast import export
from inmanta.execute.util import Unknown
from inmanta.stable_api import stable_api
from inmanta.types import DataclassProtocol
from inmanta.warnings import InmantaWarning

if TYPE_CHECKING:
    from inmanta import references
    from inmanta.ast.attribute import Attribute  # noqa: F401
    from inmanta.ast.entity import Entity
    from inmanta.ast.statements import Statement  # noqa: F401
    from inmanta.ast.statements.define import DefineEntity, DefineImport  # noqa: F401
    from inmanta.ast.type import NamedType, Type  # noqa: F401
    from inmanta.compiler import Compiler
    from inmanta.execute.runtime import DelayedResultVariable, ExecutionContext, Instance, ResultVariable  # noqa: F401


class TypeDeprecationWarning(InmantaWarning):
    pass


class Location(export.Exportable):
    __slots__ = ("file", "lnr")

    def __init__(self, file: str, lnr: int) -> None:
        self.file = file
        self.lnr = lnr

    def __str__(self) -> str:
        return "%s:%d" % (self.file, self.lnr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Location):
            return False
        return self.file == other.file and self.lnr == other.lnr

    def merge(self, other: "Location") -> "Location":
        if other is None:
            return self

        assert isinstance(other, Location)
        assert self.file == other.file

        return Location(self.file, min(self.lnr, other.lnr))

    def export(self) -> export.Location:
        # Location is 1-based, export.Position spec is 0-based
        # whole line: range from line:0 to line+1:0
        range_start: export.Position = export.Position(line=self.lnr - 1, character=0)
        range_end: export.Position = export.Position(line=self.lnr, character=0)
        return export.Location(uri=self.file, range=export.Range(start=range_start, end=range_end))


class Range(Location):
    __slots__ = ("start_char", "end_lnr", "end_char")

    def __init__(self, file: str, start_lnr: int, start_char: int, end_lnr: int, end_char: int) -> None:
        """
        Create a new Range instance.
        :param file: the file this Range is in
        :param start_lnr: the line number this Range starts on, 1-based
        :param start_char: the start character number of the Range, 1-based
        :param end_lnr: the line number this Range ends on, 1-based
        :param end_char: the end character number of the Range, exclusive, 1-based
        """
        Location.__init__(self, file, start_lnr)
        self.start_char = start_char
        self.end_lnr = end_lnr
        self.end_char = end_char

    def merge(self, other: Location) -> Location:
        if other is None:
            return self

        assert isinstance(other, Location)
        assert self.file == other.file

        if not isinstance(other, Range):
            return Location(self.file, min(self.lnr, other.lnr))
        else:
            if other.lnr < self.lnr:
                lnr = other.lnr
                start_char = other.start_char
            elif other.lnr == self.lnr:
                lnr = other.lnr
                start_char = min(self.start_char, other.start_char)
            else:
                lnr = self.lnr
                start_char = self.start_char

            if other.end_lnr > self.end_lnr:
                end_lnr = other.end_lnr
                end_char = other.end_char
            elif other.end_lnr == self.end_lnr:
                end_lnr = other.end_lnr
                end_char = max(self.end_char, other.end_char)
            else:
                end_lnr = self.lnr
                end_char = self.end_char
            return Range(self.file, lnr, start_char, end_lnr, end_char)

    def export(self) -> export.Location:
        range_start: export.Position = export.Position(line=self.lnr - 1, character=self.start_char - 1)
        range_end: export.Position = export.Position(line=self.end_lnr - 1, character=self.end_char - 1)
        result: export.Location = super().export()
        result.range = export.Range(start=range_start, end=range_end)
        return result

    def __str__(self) -> str:
        return "%s:%d:%d" % (self.file, self.lnr, self.start_char)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Range):
            return (
                self.file == other.file
                and self.lnr == other.lnr
                and self.start_char == other.start_char
                and self.end_lnr == other.end_lnr
                and self.end_char == other.end_char
            )
        return False


class AnchorTarget:
    """AnchorTarget is used purely at the periphery of the compiler"""

    __slots__ = ("location", "docstring")

    def __init__(
        self,
        location: Location,
        docstring: Optional[str] = None,
    ) -> None:
        """
        Create a new AnchorTarget instance.
        :param location: the location of the target of the anchor
        :param docstring: the docstring attached to the target
        """
        self.location = location
        self.docstring = docstring


class WithComment:
    """
    Mixin class for AST nodes that can have a comment attached to them.
    """

    comment: Optional[str] = None


class Locatable:
    __slots__ = ("_location",)

    def __init__(self) -> None:
        self._location: Optional[Location] = None

    def set_location(self, location: Location) -> None:
        assert location is not None and location.lnr > 0
        self._location = location

    def get_location(self) -> Optional[Location]:
        assert self._location is not None
        return self._location

    def copy_location(self, other: "Locatable") -> None:
        """
        Copy the location of this locatable to the given locatable
        """
        other.set_location(self.location)

    location = property(get_location, set_location)


class LocatableString:
    """
    A string with an attached source location.

    It is not a subtype of str, as str is not a normal class
    As such, it is very important to unwrap strings as this object is not an actual string.

    All identifiers produced by the parser are of this type.

    The unwrapping should be done in
    1. anywhere in DefinitionStatements
    2. in the constructors of other statements
    """

    def __init__(self, value: str, location: Range, lexpos: int, namespace: "Namespace") -> None:
        self.value = value
        self.location = location

        self.lnr = location.lnr
        self.elnr = location.end_lnr
        self.end = location.end_char
        self.start = location.start_char

        self.lexpos = lexpos
        self.namespace = namespace

    def get_value(self) -> str:
        return self.value

    def get_location(self) -> Range:
        return self.location

    def __str__(self) -> str:
        return self.value


class Anchor:
    def __init__(self, range: Range) -> None:
        self.range = range

    def get_range(self) -> Range:
        return self.range

    def get_location(self) -> Range:
        return self.range

    @abstractmethod
    def resolve(self) -> Optional[AnchorTarget]:
        raise NotImplementedError()


class TypeReferenceAnchor(Anchor):
    def __init__(self, namespace: "Namespace", type: LocatableString) -> None:
        Anchor.__init__(self, range=type.get_location())
        self.namespace = namespace
        self.type = type

    def resolve(self) -> Optional[AnchorTarget]:
        t = self.namespace.get_type(self.type)
        location = t.get_location()
        docstring = t.comment if isinstance(t, WithComment) else None
        if not location:
            return None
        return AnchorTarget(location=location, docstring=docstring)


class TypeAnchor(Anchor):
    """Reference to a resolved type"""

    def __init__(self, reference: LocatableString, type: "Type") -> None:
        """
        :param reference: the location we are referencing from
        :param type: the type that is being referenced
        """
        Anchor.__init__(self, range=reference.get_location())
        self.type = type
        self.type.get_location()

    def resolve(self) -> Optional[AnchorTarget]:
        t = self.type
        location = t.get_location()
        docstring = t.comment if isinstance(t, WithComment) else None
        if not location:
            return None
        return AnchorTarget(location=location, docstring=docstring)


class AttributeAnchor(Anchor):
    def __init__(self, range: Range, attribute: "Attribute") -> None:
        Anchor.__init__(self, range=range)
        self.attribute = attribute

    def resolve(self) -> Optional[AnchorTarget]:
        location = self.attribute.get_location()
        docstring = self.attribute.comment if isinstance(self.attribute, WithComment) else None
        if not location:
            return None
        return AnchorTarget(location=location, docstring=docstring)


class Namespaced(Locatable):
    __slots__ = ()

    @abstractmethod
    def get_namespace(self) -> "Namespace":
        raise NotImplementedError()


class Named(Namespaced):
    @abstractmethod
    def get_full_name(self) -> str:
        raise NotImplementedError()


class Import(Locatable):
    def __init__(self, target: "Namespace") -> None:
        Locatable.__init__(self)
        self.target = target


class MockImport(Import):
    def __init__(self, target: "Namespace") -> None:
        Locatable.__init__(self)
        self.target = target


class Namespace(Namespaced):
    """
    This class models a namespace that contains defined types, modules, ...
    """

    __slots__ = ("__name", "__parent", "__children", "defines_types", "visible_namespaces", "primitives", "scope", "__path")

    def __init__(self, name: str, parent: "Optional[Namespace]" = None) -> None:
        Namespaced.__init__(self)
        self.__name = name
        self.__parent = parent
        self.__children = {}  # type: Dict[str,Namespace]
        self.defines_types = {}  # type: Dict[str,NamedType]
        self.visible_namespaces: dict[str, Import]
        if self.__parent is not None:
            self.visible_namespaces = {self.get_full_name(): MockImport(self)}
            self.__parent.add_child(self)
        else:
            self.visible_namespaces = {name: MockImport(self)}
        self.primitives = None  # type: Optional[Dict[str,Type]]
        self.scope = None  # type:  Optional[ExecutionContext]
        self.__path: list[str]

        if self.__parent is None or self.__parent.get_name() == "__root__":
            self.__path = [self.__name]
        else:
            self.__path = self.__parent.to_path() + [self.__name]

    def set_primitives(self, primitives: "Dict[str,Type]") -> None:
        self.primitives = primitives
        for child in self.children():
            child.set_primitives(primitives)

        std = self.get_ns_from_string("std")
        assert std is not None

        self.visible_namespaces["std"] = MockImport(std)

    def get_primitives(self) -> "Dict[str,Type]":
        assert self.primitives is not None
        return self.primitives

    def get_scope(self) -> "ExecutionContext":
        assert self.scope is not None
        return self.scope

    def define_type(self, name: str, newtype: "NamedType") -> None:
        if name in self.defines_types:
            raise newtype.get_double_defined_exception(self.defines_types[name])
        self.defines_types[name] = newtype

    def import_ns(self, name: str, ns: "DefineImport") -> None:
        if name in self.visible_namespaces:
            other = self.visible_namespaces[name]
            if not isinstance(other, MockImport):
                raise DuplicateException(ns, self.visible_namespaces[name], "Two import statements have the same name")
        self.visible_namespaces[name] = ns

    def lookup_namespace(self, name: str) -> Import:
        if name not in self.visible_namespaces:
            raise NotFoundException(None, name, f"Namespace {name} not found.\nTry importing it with `import {name}`")
        return self.visible_namespaces[name]

    def lookup(self, name: str) -> "Union[Type, ResultVariable]":
        if "::" not in name:
            return self.get_scope().direct_lookup(name)
        parts = name.rsplit("::", 1)
        return self.lookup_namespace(parts[0]).target.get_scope().direct_lookup(parts[1])

    def get_type(self, typ: LocatableString) -> "Type":
        name: str = str(typ)
        assert self.primitives is not None
        if "::" in name:
            parts = name.rsplit("::", 1)
            if parts[0] in self.visible_namespaces:
                ns = self.visible_namespaces[parts[0]].target
                if parts[1] in ns.defines_types:
                    return ns.defines_types[parts[1]]
                else:
                    raise TypeNotFoundException(typ, ns)
            else:
                raise MissingImportException(typ, self, parts, str(self.location.file))
        elif name in self.primitives:
            if name == "number":
                warnings.warn(TypeDeprecationWarning("Type 'number' is deprecated, use 'float' or 'int' instead"))
            return self.primitives[name]
        else:
            cns = self  # type: Optional[Namespace]
            while cns is not None:
                if name in cns.defines_types:
                    return cns.defines_types[name]
                cns = cns.get_parent()

            raise TypeNotFoundException(typ, self)

    def get_name(self) -> str:
        """
        Get the name of this namespace
        """
        return self.__name

    def get_full_name(self) -> str:
        """
        Get the fully qualified name of this namespace
        """
        if self.__parent is None:
            raise Exception("Should not occur, compiler corrupt")
        if self.__parent.__parent is None:
            return self.get_name()
        return self.__parent.get_full_name() + "::" + self.get_name()

    name = property(get_name)

    def set_parent(self, parent: "Namespace") -> None:
        """
        Set the parent of this namespace. This namespace is also added to
        the child list of the parent.
        """
        self.__parent = parent
        self.__parent.add_child(self)

    def get_parent(self) -> "Optional[Namespace]":
        """
        Get the parent namespace
        """
        return self.__parent

    parent = property(get_parent, set_parent)

    def get_root(self) -> "Namespace":
        """
        Get the root
        """
        if self.__parent is None:
            return self
        return self.__parent.get_root()

    def add_child(self, child_ns: "Namespace") -> None:
        """
        Add a child to the namespace.
        """
        self.__children[child_ns.get_name()] = child_ns

    def __repr__(self) -> str:
        """
        The representation of this object
        """
        if self.__parent is not None and self.__parent.get_name() != "__root__":
            return repr(self.__parent) + "::" + self.__name

        return self.__name

    def children(self, recursive: bool = False) -> "List[Namespace]":
        """
        Get the children of this namespace
        """
        children = list(self.__children.values())
        if not recursive:
            return children

        for child in self.__children.values():
            children.extend(child.children(recursive=True))

        return children

    def get_child(self, name: str) -> "Optional[Namespace]":
        """
        Returns the child namespace with the given name or None if it does
        not exist.
        """
        if name in self.__children:
            return self.__children[name]
        return None

    def get_child_or_create(self, name: str) -> "Namespace":
        """
        Returns the child namespace with the given name or None if it does
        not exist.
        """
        if name in self.__children:
            return self.__children[name]
        out = Namespace(name, self)
        return out

    def get_ns_or_create(self, name: str) -> "Namespace":
        """
        Returns the child namespace with the given name or None if it does
        not exist.
        """
        name_parts = name.split("::")
        if len(name_parts) == 1:
            parent = self.get_root()
        else:
            preparent = self.get_root()._get_ns(name_parts[:-1])
            assert preparent is not None
            parent = preparent
        return parent.get_child_or_create(name_parts[-1])

    def get_ns_from_string(self, fqtn: str) -> "Optional[Namespace]":
        """
        Get the namespace that is referenced to in the given fully qualified
        type name.

        :param fqtn: The type name
        """
        name_parts = fqtn.split("::")
        return self.get_root()._get_ns(name_parts)

    def _get_ns(self, ns_parts: list[str]) -> "Optional[Namespace]":
        """
        Return the namespace indicated by the parts list. Each element of
        the array represents a level in the namespace hierarchy.
        """
        if len(ns_parts) == 0:
            return None
        elif len(ns_parts) == 1:
            return self.get_child(ns_parts[0])
        else:
            child = self.get_child(ns_parts[0])
            if child is None:
                return None
            return child._get_ns(ns_parts[1:])

    def to_path(self) -> list[str]:
        """
        Return a list with the namespace path elements in it.
        """
        return self.__path

    def get_namespace(self) -> "Namespace":
        return self

    def get_location(self) -> Location:
        return self.location


@stable_api
class CompilerException(Exception, export.Exportable):
    """Base class for exceptions generated by the compiler"""

    def __init__(self, msg: str) -> None:
        Exception.__init__(self, msg)
        self.location = None  # type: Optional[Location]
        self.msg = msg
        # store root namespace so error reporters can inspect the compiler state
        self.root_ns: Optional[Namespace] = None

    def set_location(self, location: Location) -> None:
        if self.location is None:
            self.location = location

    def get_message(self) -> str:
        return self.msg

    def get_location(self) -> Optional[Location]:
        return self.location

    def get_causes(self) -> "List[CompilerException]":
        return []

    def format(self) -> str:
        """Make a string representation of this particular exception"""
        location = self.get_location()
        if location is not None:
            return f"{self.get_message()} ({location})"
        else:
            return self.get_message()

    def format_trace(self, indent: str = "", indent_level: int = 0) -> str:
        """Make a representation of this exception and its causes"""
        out = indent * indent_level + self.format()

        for cause in self.get_causes():
            part = cause.format_trace(indent=indent, indent_level=indent_level + 1)
            out += "\n" + indent * indent_level + "caused by:"
            out += "\n" + part

        return out

    def importantance(self) -> int:
        """
        Importance used to order exceptions when reporting multiple, lower is more important

        default is 100
        below 50 is for pure compiler errors (type, syntax)
        """
        return 100

    def attach_compile_info(self, compiler: "Compiler") -> None:
        self.root_ns = compiler.get_ns()

    def export(self) -> export.Error:
        location: Optional[Location] = self.get_location()
        module: Optional[str] = self.__class__.__module__
        name: str = self.__class__.__qualname__
        return export.Error(
            type=name if module is None else f"{module}.{name}",
            message=self.get_message(),
            location=location.export() if location is not None else None,
        )

    def __str__(self) -> str:
        return self.format()


@stable_api
class RuntimeException(CompilerException):
    """Baseclass for exceptions raised by the compiler after parsing is complete."""

    def __init__(self, stmt: "Optional[Locatable]", msg: str) -> None:
        CompilerException.__init__(self, msg)
        self.stmt = None
        if stmt is not None:
            self.set_location(stmt.get_location())
            self.stmt = stmt

    def set_statement(self, stmt: "Locatable", replace: bool = True) -> None:
        for cause in self.get_causes():
            if isinstance(cause, RuntimeException):
                cause.set_statement(stmt, replace)

        if replace or self.stmt is None:
            self.set_location(stmt.get_location())
            self.stmt = stmt

    def format(self) -> str:
        """Make a string representation of this particular exception"""
        if self.stmt is not None:
            return f"{self.get_message()} (reported in {self.stmt} ({self.get_location()}))"
        return super().format()


class InvalidCompilerState(RuntimeException):
    def __init__(self, stmt: "Optional[Locatable]", msg: str) -> None:
        super().__init__(stmt, "Invalid compiler state, this likely indicates a bug in the compiler. Details: %s" % msg)


class HyphenException(RuntimeException):
    def __init__(self, stmt: LocatableString) -> None:
        msg: str = "The use of '-' in identifiers is not allowed. please rename %s." % stmt.value
        RuntimeException.__init__(self, stmt, msg)


class CompilerRuntimeWarning(InmantaWarning, RuntimeException):
    """
    Baseclass for compiler warnings after parsing is complete.
    """

    def __init__(self, stmt: "Optional[Locatable]", msg: str) -> None:
        InmantaWarning.__init__(self)
        RuntimeException.__init__(self, stmt, msg)


class CompilerDeprecationWarning(CompilerRuntimeWarning):
    def __init__(self, stmt: Optional["Locatable"], msg: str) -> None:
        CompilerRuntimeWarning.__init__(self, stmt, msg)


class VariableShadowWarning(CompilerRuntimeWarning):
    def __init__(self, stmt: Optional["Locatable"], msg: str):
        CompilerRuntimeWarning.__init__(self, stmt, msg)


class TypeNotFoundException(RuntimeException):
    """Exception raised when a type is referenced that does not exist"""

    def __init__(self, type: LocatableString, ns: Namespace) -> None:
        RuntimeException.__init__(self, stmt=None, msg=f"could not find type {type} in namespace {ns}")
        self.type = type
        self.ns = ns
        self.set_location(type.get_location())

    def importantance(self) -> int:
        return 20


class MissingImportException(TypeNotFoundException):
    """Exception raised when a referenced type's module is missing from the namespace"""

    def __init__(
        self, type: LocatableString, ns: Namespace, parts: Optional[list[str]] = None, file: Optional[str] = None
    ) -> None:
        suggest_importing: str = ""
        suggest_file = ""
        if file:
            suggest_file = f" in {file}"
        if parts:
            suggest_importing = f".\nTry importing the module with `import {parts[0]}`{suggest_file}"
        super().__init__(type, ns)
        self.msg += suggest_importing


class AmbiguousTypeException(TypeNotFoundException):
    """
    Exception raised when a type is referenced in an unqualified manner in a context where it can not be unambiguously resolved.
    """

    def __init__(self, type: LocatableString, candidates: list["Entity"]) -> None:
        candidates = sorted(candidates, key=lambda x: x.get_full_name())
        RuntimeException.__init__(
            self,
            stmt=None,
            msg="Could not determine namespace for type %s. %d possible candidates exists: [%s]."
            " To resolve this, use the fully qualified name instead of the short name."
            % (type, len(candidates), ", ".join([x.get_full_name() for x in candidates])),
        )
        self.candidates = candidates
        self.type = type
        self.set_location(type.get_location())


def stringify_exception(exn: Exception) -> str:
    if isinstance(exn, CompilerException):
        return str(exn)
    return f"{exn.__class__.__name__}: {exn}"


@stable_api
class ExternalException(RuntimeException):
    """
    When a plugin call produces an exception that is not a :py:class:`RuntimeException`,
    it is wrapped in an ExternalException to make it conform to the expected interface
    """

    def __init__(self, stmt: Optional[Locatable], msg: str, cause: Exception) -> None:
        RuntimeException.__init__(self, stmt=stmt, msg=msg)

        self.__cause__ = cause

    def get_causes(self) -> list[CompilerException]:
        return []

    def format_trace(self, indent: str = "", indent_level: int = 0) -> str:
        """Make a representation of this exception and its causes"""

        out = indent * indent_level + self.format().replace("\n", "\n" + indent * indent_level)

        part = traceback.format_exception_only(self.__cause__.__class__, self.__cause__)
        out += "\n" + indent * indent_level + "caused by:\n"
        for line in part:
            out += indent * (indent_level + 1) + line

        return out

    def importantance(self) -> int:
        return 60


@stable_api
class ExplicitPluginException(ExternalException):
    """
    Base exception for wrapping an explicit :py:class:`inmanta.plugins.PluginException` raised from a plugin call.
    """

    def __init__(self, stmt: "Optional[Locatable]", msg: str, cause: "PluginException") -> None:
        ExternalException.__init__(self, stmt, msg, cause)
        self.__cause__: PluginException

    def export(self) -> export.Error:
        location: Optional[Location] = self.get_location()
        module: Optional[str] = self.__cause__.__class__.__module__
        name: str = self.__cause__.__class__.__qualname__
        return export.Error(
            type=name if module is None else f"{module}.{name}",
            message=self.__cause__.message,
            location=location.export() if location is not None else None,
            category=export.ErrorCategory.plugin,
        )

    def get_message(self) -> str:
        return self.msg + "\n" + self.__cause__.message


class WrappingRuntimeException(RuntimeException):
    """Baseclass for RuntimeExceptions wrapping other CompilerException"""

    def __init__(self, stmt: "Optional[Locatable]", msg: str, cause: CompilerException) -> None:
        if stmt is None and isinstance(cause, RuntimeException):
            stmt = cause.stmt

        RuntimeException.__init__(self, stmt=stmt, msg=msg)

        self.__cause__ = cause  # type: CompilerException

    def get_causes(self) -> list[CompilerException]:
        return [self.__cause__]

    def importantance(self) -> int:
        # less likely to be the cause then out child
        return self.__cause__.importantance() + 1


@stable_api
class AttributeException(WrappingRuntimeException):
    """Exception raise when an attribute could not be set, always wraps another exception"""

    def __init__(self, stmt: "Locatable", instance: "Instance", attribute: str, cause: RuntimeException) -> None:
        WrappingRuntimeException.__init__(
            self, stmt=stmt, msg=f"Could not set attribute `{attribute}` on instance `{instance}`", cause=cause
        )
        self.attribute = attribute
        self.instance = instance


class PluginTypeException(WrappingRuntimeException):
    """
    Raised when an argument or return value of a plugin is not in-line with the type
    annotation of that plugin. Wraps the validation error.
    """

    pass


class OptionalValueException(RuntimeException):
    """Exception raised when an optional value is accessed that has no value (and is frozen)"""

    def __init__(self, instance: "Instance", attribute: "Attribute") -> None:
        RuntimeException.__init__(
            self, None, f"Optional variable accessed that has no value (attribute `{attribute}` of `{instance}`)"
        )
        self.instance = instance
        self.attribute = attribute

    def importantance(self) -> int:
        return 61


class IndexException(RuntimeException):
    """Exception raised when an index definition is invalid"""

    def importantance(self) -> int:
        return 10


class TypingException(RuntimeException):
    """Base class for exceptions raised during the typing phase of compilation"""

    def importantance(self) -> int:
        return 10


class InvalidTypeAnnotation(TypingException):
    """
    Invalid type annotation, e.g. for a plugin.
    """


class DirectExecuteException(TypingException):
    """Exception raised when direct execute is called on a wrong object"""

    def importantance(self) -> int:
        return 11


class DataClassException(TypingException):
    """Exception in relation to dataclasses"""

    def __init__(self, entity: "Entity", msg: str) -> None:
        super().__init__(entity, msg)
        self.entity = entity

    def importantance(self) -> int:
        return 9


class DataClassMismatchException(DataClassException):
    """
    Exception due to a mismatch between both version of a dataclass

    """

    def __init__(
        self,
        entity: "Entity",
        dataclass: DataclassProtocol | None,
        dataclass_python_name: str,
        msg: str,
    ) -> None:
        """
        :param dataclass python dataclass, None if absent
        """
        super().__init__(entity, msg)
        self.dataclass = dataclass
        self.dataclass_python_name = dataclass_python_name


class KeyException(RuntimeException):
    pass

    def importantance(self) -> int:
        return 70


class CycleException(TypingException):
    """Exception raised when a type is its own parent (type cycle)"""

    def __init__(self, first_type: "DefineEntity", final_name: str) -> None:
        super().__init__(first_type, "")
        self.types = []  # type: List[DefineEntity]
        self.complete = False
        self.final_name = final_name

    def add(self, element: "DefineEntity") -> None:
        """Collect parent entities while traveling up the stack"""
        if self.complete:
            return
        if element.get_full_name() == self.final_name:
            self.complete = True
        self.types.append(element)

    def get_message(self) -> str:
        trace = ",".join([x.get_full_name() for x in self.types])
        return "Entity can not be its own parent %s" % (trace)


class NotFoundException(RuntimeException):
    def __init__(self, stmt: "Optional[Statement]", name: str, msg: "Optional[str]" = None) -> None:
        if msg is None:
            msg = "could not find value %s" % name
        RuntimeException.__init__(self, stmt, msg)
        self.name = name

    def importantance(self) -> int:
        return 20


@stable_api
class DoubleSetException(RuntimeException):
    def __init__(
        self, variable: "ResultVariable", stmt: "Optional[Statement]", newvalue: object, newlocation: Location
    ) -> None:
        self.variable: "ResultVariable" = variable
        self.newvalue = newvalue  # type: object
        self.newlocation = newlocation
        msg = "value set twice:\n\told value: {}\n\t\tset at {}\n\tnew value: {}\n\t\tset at {}\n".format(
            self.variable.value,
            self.variable.location,
            self.newvalue,
            self.newlocation,
        )
        RuntimeException.__init__(self, stmt, msg)

    def importantance(self) -> int:
        return 51


class ModifiedAfterFreezeException(RuntimeException):
    def __init__(
        self,
        rv: "DelayedResultVariable",
        instance: "Instance",
        attribute: "Attribute",
        value: object,
        location: Location,
        reverse: bool,
    ) -> None:
        RuntimeException.__init__(self, None, "List modified after freeze")
        self.instance = instance
        self.attribute = attribute
        self.value = value
        self.location = location
        self.resultvariable = rv
        self.reverse = reverse

    def importantance(self) -> int:
        return 50


class DuplicateException(TypingException):
    """Exception raise when something is defined twice"""

    def __init__(self, stmt: Locatable, other: Locatable, msg: str) -> None:
        TypingException.__init__(self, stmt, msg)
        self.other = other

    def format(self) -> str:
        return f"{self.get_message()} (original at ({self.location})) (duplicate at ({self.other.get_location()}))"

    def importantance(self) -> int:
        return 40


class CompilerError(Exception):
    pass


class MultiException(CompilerException):
    """A single exception collecting multiple CompilerExceptions"""

    def __init__(self, others: list[CompilerException]) -> None:
        CompilerException.__init__(self, "")
        self.others = others

    def get_causes(self) -> list[CompilerException]:
        def sortkey(item: CompilerException):
            location = item.get_location()
            if not location:
                file = ""
                line = 0
            else:
                file = location.file
                line = location.lnr

            return (item.importantance(), file, line)

        return sorted(self.others, key=sortkey)

    def format(self) -> str:
        return "Reported %d errors" % len(self.others)

    def __str__(self) -> str:
        return "Reported %d errors:\n\t" % len(self.others) + "\n\t".join([str(e) for e in self.others])

    def format_trace(self, indent: str = "", indent_level: int = 0) -> str:
        """Make a representation of this exception and its causes"""
        out = indent * indent_level + self.format()

        for i, cause in enumerate(self.get_causes()):
            part = cause.format_trace(indent=indent, indent_level=indent_level + 1)
            out += "\n" + indent * indent_level + f"error {i}:"
            out += "\n" + part

        return out


class UnsetException(RuntimeException):
    """
    This exception is thrown when an attribute is accessed that was not yet
    available (i.e. it has not been frozen yet).
    """

    def __init__(self, msg: str, instance: Optional["Instance"] = None, attribute: Optional["Attribute"] = None) -> None:
        RuntimeException.__init__(self, None, msg)
        self.instance: Optional["Instance"] = instance
        self.attribute: Optional["Attribute"] = attribute
        self.msg = msg

    def get_result_variable(self) -> Optional["Instance"]:
        return self.instance


class MultiUnsetException(RuntimeException):
    """
    This exception is thrown when multiple attributes are accessed that were not yet
    available (i.e. it has not been frozen yet).
    """

    def __init__(self, msg: str, result_variables: "list[ResultVariable[object]]") -> None:
        RuntimeException.__init__(self, None, msg)
        self.result_variables = result_variables


class UnknownException(Exception):
    """
    This exception is thrown when code tries to access a value that is
    unknown and cannot be determined during this evaluation. The code
    receiving this exception is responsible for invalidating any results
    depending on this value by return an instance of Unknown as well.
    """

    def __init__(self, unknown: Unknown):
        super().__init__()
        self.unknown = unknown


class AttributeNotFound(NotFoundException, AttributeError):
    """
    Exception used for backwards compatibility with try-except blocks around some_proxy.some_attr.
    This previously raised `NotFoundException` which is currently deprecated in this context.
    Its new behavior is to raise an AttributeError for compatibility with Python's builtin `hasattr`.
    """


# custom class to enable clean wrapping on the plugin boundary
class UnexpectedReference(RuntimeException):
    """
    Unexpected (undeclared) reference encountered during compiler or plugin execution.
    """

    def __init__(
        self, *, stmt: Optional[Locatable] = None, reference: "references.Reference[references.RefValue]", message: str
    ) -> None:
        RuntimeException.__init__(self, stmt, message)
        self.reference: references.Reference[references.RefValue] = reference


@stable_api
class PluginException(Exception):
    """
    Base class for custom exceptions raised from a plugin.
    """

    def __init__(self, message: str) -> None:
        self.message = message



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import collections.abc
import dataclasses
import functools
import inspect
import logging
import numbers
import os
import subprocess
import typing
import warnings
from abc import abstractmethod
from collections import abc
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, Self, Type, TypeVar

import typing_inspect

import inmanta.ast.type as inmanta_type
from inmanta import const, protocol, util
from inmanta.ast import InvalidTypeAnnotation, LocatableString, Location, MultiUnsetException, Namespace
from inmanta.ast import PluginException as PluginException  # noqa: F401 Plugin exception is part of the stable api
from inmanta.ast import (
    PluginTypeException,
    Range,
    RuntimeException,
    TypeNotFoundException,
    TypingException,
    UnexpectedReference,
    UnsetException,
    WithComment,
)
from inmanta.ast.type import NamedType
from inmanta.ast.type import Null as Null  # Moved, part of stable api
from inmanta.ast.type import ReferenceType
from inmanta.config import Config
from inmanta.execute import proxy
from inmanta.execute.proxy import DynamicProxy
from inmanta.execute.runtime import QueueScheduler, Resolver, ResultVariable
from inmanta.execute.util import NoneValue, Unknown
from inmanta.references import Reference
from inmanta.stable_api import stable_api
from inmanta.warnings import InmantaWarning

T = TypeVar("T")
T_FUNC = TypeVar("T_FUNC", bound=Callable[..., object])

if TYPE_CHECKING:
    from inmanta.ast.entity import Entity
    from inmanta.ast.statements import DynamicStatement
    from inmanta.ast.statements.call import FunctionCall
    from inmanta.compiler import Compiler


LOGGER = logging.getLogger(__name__)


class PluginDeprecationWarning(InmantaWarning):
    pass


@stable_api
class Context:
    """
    An instance of this class is used to pass context to the plugin
    """

    __client: Optional["protocol.Client"] = None
    __sync_client = None

    @classmethod
    def __get_client(cls) -> "protocol.Client":
        if cls.__client is None:
            cls.__client = protocol.Client("compiler")
        return cls.__client

    def __init__(
        self, resolver: Resolver, queue: QueueScheduler, owner: "FunctionCall", plugin: "Plugin", result: ResultVariable
    ) -> None:
        self.resolver = resolver
        self.queue = queue
        self.owner = owner
        self.plugin = plugin
        self.result = result
        self.compiler = queue.get_compiler()

    def get_resolver(self) -> Resolver:
        return self.resolver

    def get_type(self, name: LocatableString) -> inmanta_type.Type:
        """
        Get a type from the configuration model.
        """
        try:
            return self.queue.get_types()[str(name)]
        except KeyError:
            raise TypeNotFoundException(name, self.owner.namespace)

    def get_queue_scheduler(self) -> QueueScheduler:
        return self.queue

    def get_environment_id(self) -> str:
        env = str(Config.get("config", "environment", None))

        if env is None:
            raise Exception("The environment of the model should be configured in config>environment")

        return env

    def get_compiler(self) -> "Compiler":
        return self.queue.get_compiler()

    def get_data_dir(self) -> str:
        """
        Get the path to the data dir (and create if it does not exist yet
        """
        data_dir = os.path.join("data", self.plugin.namespace.get_full_name())

        if not os.path.exists(data_dir):
            os.makedirs(data_dir, exist_ok=True)

        return data_dir

    def get_client(self) -> "protocol.Client":
        return self.__class__.__get_client()

    def get_sync_client(self) -> "protocol.SyncClient":
        if self.__class__.__sync_client is None:
            self.__class__.__sync_client = protocol.SyncClient("compiler")
        return self.__class__.__sync_client

    def run_sync(self, function: Callable[[], abc.Awaitable[T]], timeout: int = 5) -> T:
        """
        Execute the async function and return its result. This method uses this thread's current (not running) event loop if
        there is one, otherwise it creates a new one. The main use for this function is to use the inmanta internal rpc to
        communicate with the server.

        :param function: The async function to execute. This function should return a yieldable object.
        :param timeout: A timeout for the async function.
        :return: The result of the async call.
        :raises ConnectionRefusedError: When the function timeouts this exception is raised.
        """
        try:
            return util.wait_sync(function(), timeout=timeout)
        except TimeoutError:
            raise ConnectionRefusedError()


@stable_api
class PluginMeta(type):
    """
    A metaclass that keeps track of concrete plugin subclasses. This class is responsible for all plugin registration.
    """

    def __new__(cls, name: str, bases: tuple[type, ...], dct: dict[str, object]) -> type:
        subclass = type.__new__(cls, name, bases, dct)
        if hasattr(subclass, "__function_name__"):
            cls.add_function(subclass)
        return subclass

    __functions: dict[str, type["Plugin"]] = {}

    @classmethod
    def add_function(cls, plugin_class: type["Plugin"]) -> None:
        """
        Add a function plugin class
        """
        cls.__functions[plugin_class.__fq_plugin_name__] = plugin_class

    @classmethod
    def get_functions(cls) -> dict[str, "Type[Plugin]"]:
        """
        Get all functions that are registered
        """
        return dict(cls.__functions)

    @classmethod
    def clear(cls, inmanta_module: Optional[str] = None) -> None:
        """
        Clears registered plugin functions.

        :param inmanta_module: Clear plugin functions for a specific inmanta module. If omitted, clears all registered plugin
            functions.
        """
        if inmanta_module is not None:
            top_level: str = f"{const.PLUGINS_PACKAGE}.{inmanta_module}"
            cls.__functions = {
                fq_name: plugin_class
                for fq_name, plugin_class in cls.__functions.items()
                if plugin_class.__module__ != top_level and not plugin_class.__module__.startswith(f"{top_level}.")
            }
        else:
            cls.__functions = {}


class UnConvertibleEntity(inmanta_type.Type):
    """
    Entity that does not convert to a dataclass.
    """

    def __init__(self, base_entity: "Entity") -> None:
        super().__init__()
        self.base_entity = base_entity

    def validate(self, value: Optional[object]) -> bool:
        return self.base_entity.validate(value)

    def type_string(self) -> Optional[str]:
        return self.base_entity.type_string()

    def type_string_internal(self) -> str:
        return self.base_entity.type_string_internal()

    def normalize(self) -> None:
        pass

    def is_attribute_type(self) -> bool:
        return False

    def get_base_type(self) -> "inmanta_type.Type":
        return self.base_entity.get_base_type()

    def with_base_type(self, base_type: "inmanta_type.Type") -> "inmanta_type.Type":
        return self.base_entity.with_base_type(base_type)

    def corresponds_to(self, type: "inmanta_type.Type") -> bool:
        raise NotImplementedError()

    def as_python_type_string(self) -> "str | None":
        return self.base_entity.as_python_type_string()

    def has_custom_to_python(self) -> bool:
        return False

    def to_python(self, instance: object, *, path: str) -> object:
        return self.base_entity.to_python(instance, path=path)

    def get_location(self) -> Optional[Location]:
        return self.base_entity.get_location()


# Define some types which are used in the context of plugins.
PLUGIN_TYPES = {
    "any": inmanta_type.Any(),  # Any value will pass validation
    "expression": inmanta_type.Any(),  # Any value will pass validation
    "null": Null(),  # Only NoneValue will pass validation
    None: Null(),  # Only NoneValue will pass validation
}

python_to_model = {
    str: inmanta_type.String(),
    float: inmanta_type.Float(),
    numbers.Number: inmanta_type.Number(),
    int: inmanta_type.Integer(),
    bool: inmanta_type.Bool(),
    dict: inmanta_type.TypedDict(inmanta_type.Any()),
    typing.Mapping: inmanta_type.TypedDict(inmanta_type.Any()),
    Mapping: inmanta_type.TypedDict(inmanta_type.Any()),
    list: inmanta_type.List(),
    typing.Sequence: inmanta_type.List(),
    Sequence: inmanta_type.List(),
    object: inmanta_type.Any(),
    # unannotated Reference -> Reference[object]
    Reference: ReferenceType(inmanta_type.Any()),
}


@dataclass(frozen=True)
class ModelType:
    """
    Dataclass used with typing.Annotated to represent Inmanta model types in Python code.

    If we want to represent "std::Entity" as "typing.Any" in our plugins we could define the following type:

    type Entity = typing.Annotated[typing.Any, ModelType["std::Entity"]]

    and then use it on our plugin:

    @plugin
    def my_plugin(value: Entity) -> None:
        pass

    We will validate the argument as "std::Entity", while presenting
    a proper Python type (typing.Any) for IDE and static typing purposes.
    It is the user's responsibility to ensure that the validation type sufficiently matches the Python type.

    :param model_type: The fully qualified name of the Inmanta model type
    """

    model_type: str

    def __class_getitem__(cls: type[Self], key: str) -> Self:
        return cls(key)


def parse_dsl_type(dsl_type: str, location: Range, resolver: Namespace) -> inmanta_type.Type:
    if dsl_type in PLUGIN_TYPES:
        return PLUGIN_TYPES[dsl_type]
    locatable_type: LocatableString = LocatableString(dsl_type, location, 0, resolver)
    return inmanta_type.resolve_type(locatable_type, resolver)


def _convert_to_reference(
    python_type: type[object], origin: type[object], location: Range, resolver: Namespace
) -> inmanta_type.Type | None:
    if issubclass(origin, Reference):
        # We rely on the order of argument because of
        # https://github.com/ilevkivskyi/typing_inspect/issues/110
        # We can only handle the case where T is a concrete type, not where it is a re-mapped type-var
        # https://github.com/inmanta/inmanta-core/issues/8765
        args = typing.get_args(python_type)
        return ReferenceType(to_dsl_type(args[0], location, resolver))
    return None


def _convert_origin_to_dsl_type(
    python_type: type[object], origin: type[object], location: Range, resolver: Namespace
) -> inmanta_type.Type | None:
    """
    Take a `python_type` of the form `origin[args]` and try to convert it
    """
    # dict
    if issubclass(origin, Mapping):
        if origin in [collections.abc.Mapping, dict, typing.Mapping]:
            args = typing_inspect.get_args(python_type)
            if not args:
                return inmanta_type.TypedDict(inmanta_type.Any())

            if not issubclass(args[0], str):
                raise TypingException(
                    None, f"invalid type {python_type}, the keys of any dict should be 'str', got {args[0]} instead"
                )

            if len(args) == 1:
                return inmanta_type.TypedDict(inmanta_type.Any())

            return inmanta_type.TypedDict(to_dsl_type(args[1], location, resolver))
        else:
            raise TypingException(None, f"invalid type {python_type}, dictionary types should be Mapping or dict")

    # List
    if issubclass(origin, Sequence):
        if origin in [collections.abc.Sequence, list, typing.Sequence]:
            sargs = typing.get_args(python_type)
            if not sargs:
                return inmanta_type.List()
            return inmanta_type.TypedList(to_dsl_type(sargs[0], location, resolver))
        else:
            raise TypingException(None, f"invalid type {python_type}, list types should be Sequence or list")

    # Set
    if issubclass(origin, collections.abc.Set):
        raise TypingException(None, f"invalid type {python_type}, set is not supported on the plugin boundary")

    return _convert_to_reference(python_type, origin, location, resolver)


def to_dsl_type(python_type: type[object], location: Range, resolver: Namespace) -> inmanta_type.Type:
    """
    Convert a python type annotation to an Inmanta DSL type annotation.

    :param python_type: The evaluated python type as provided in the Python type annotation.
    :param location: The location of this evaluation on the model
    :param resolver: The namespace that can be used to resolve the type annotation of this argument.
    """
    # Resolve aliases
    if isinstance(python_type, typing.TypeAliasType):
        return to_dsl_type(python_type.__value__, location, resolver)

    # Any to any
    if python_type is typing.Any:
        return inmanta_type.Any()

    # None to None
    if python_type is type(None) or python_type is None:
        return Null()

    # Unions and optionals
    if typing_inspect.is_union_type(python_type):
        # Optional type
        bases: Sequence[inmanta_type.Type]
        if typing_inspect.is_optional_type(python_type):
            other_types = [tt for tt in typing.get_args(python_type) if not typing_inspect.is_optional_type(tt)]
            if len(other_types) == 0:
                # Probably not possible
                return Null()
            if len(other_types) == 1:
                return inmanta_type.NullableType(to_dsl_type(other_types[0], location, resolver))
            return inmanta_type.create_union([to_dsl_type(arg, location, resolver) for arg in other_types] + [Null()])
        else:
            bases = [to_dsl_type(arg, location, resolver) for arg in typing.get_args(python_type)]
            return inmanta_type.create_union(bases)

    if dataclasses.is_dataclass(python_type):
        entity = proxy.get_inmanta_type_for_dataclass(python_type)
        if entity:
            return entity
        raise TypingException(None, f"invalid type {python_type}, this dataclass has no associated inmanta entity")

    if dataclasses.is_dataclass(python_type):
        entity = proxy.get_inmanta_type_for_dataclass(python_type)
        if entity:
            return entity
        raise TypingException(None, f"invalid type {python_type}, this dataclass has no associated inmanta entity")

    # Lists and dicts
    if typing_inspect.is_generic_type(python_type):
        origin = typing.get_origin(python_type)
        if origin is not None:
            out = _convert_origin_to_dsl_type(python_type, origin, location, resolver)
            if out is not None:
                return out
        else:
            # We are not of the form Reference[T] but possibly a class that inherits from it
            # We do a best effort here to untangle this, but it is difficult because of
            # https://github.com/ilevkivskyi/typing_inspect/issues/110
            # We can only handle the case where T is a concrete type, not where it is a re-mapped type-var
            # https://github.com/inmanta/inmanta-core/issues/8765
            all_bases = list(typing_inspect.get_generic_bases(python_type))
            seen = set()
            while all_bases:
                base = all_bases.pop()
                # prevent loops
                if base in seen:
                    continue
                seen.add(base)

                if not typing_inspect.is_generic_type(base):
                    # Not generic, not interesting
                    continue

                origin = typing.get_origin(base)
                if origin is None:
                    # no origin, see if we have any other bases higher up
                    all_bases.extend(typing_inspect.get_generic_bases(base))
                    continue

                out = _convert_to_reference(base, origin, location, resolver)
                if out is not None:
                    return out

        # Annotated
        if origin is typing.Annotated:
            for meta in reversed(python_type.__metadata__):  # type: ignore
                if isinstance(meta, ModelType):
                    dsl_type = parse_dsl_type(meta.model_type, location, resolver)
                    # override for specific case of a dataclass: we don't want to convert
                    # correct typing is difficult due to import loop, see dsl_type.is_entity()
                    if typing.get_args(python_type)[0] is DynamicProxy and dsl_type.is_entity():
                        return UnConvertibleEntity(dsl_type)
                    return dsl_type

            # the annotation doesn't concern us => use base type
            return to_dsl_type(typing.get_args(python_type)[0], location, resolver)

    if python_type in python_to_model:
        return python_to_model[python_type]

    warnings.warn(
        InmantaWarning(
            f"Python type {python_type} was implicitly cast to 'Any' because no matching type was found in the Inmanta DSL. "
            f"Please refer to the documentation for an overview of supported types at the plugin boundary."
        )
    )

    return inmanta_type.Any()


def validate_and_convert_to_python_domain(*, name: str, expected_type: inmanta_type.Type, value: object) -> object:
    """
    Given a model domain value and an inmanta type, produce the corresponding python object

    Unknowns are not handled by this method!
    """
    expected_type.validate(value)

    if isinstance(value, NoneValue):
        # if the type is not nullable, it will fail when validating
        # if the value is None, it becomes None
        return None

    if expected_type.has_custom_to_python():
        return expected_type.to_python(value, path=name)

    return DynamicProxy.return_value(value, context=proxy.ProxyContext(path=name, validated=True))


class PluginCallContext:
    """
    Internal state of a plugin call

    Used to carry state from the argument validation to the return validation

    """


class PluginValue:
    """
    Base class for all values that go in and out of a plugin: arguments and return value.

    The class has two class attributes that should be set in the different subclasses:
    :attr VALUE_TYPE: The type of io value it is, argument or return value
    :attr VALUE_NAME: The name of the io value, the argument name or "return value"

    These attributes are only used for better error reporting.
    """

    VALUE_TYPE: str = ""
    VALUE_NAME: str = ""

    def __init__(self, type_expression: object) -> None:
        self.type_expression = type_expression
        self._resolved_type: Optional[inmanta_type.Type] = None

    @property
    def resolved_type(self) -> inmanta_type.Type:
        """
        Get the resolved type of this plugin io.  The resolved type can only be accessed
        once this object has been normalized (which happens during the plugin normalization).
        """
        if self._resolved_type is None:
            raise RuntimeException(
                stmt=None,
                msg=f"{type(self).__name__} {self.VALUE_NAME} ({repr(self.type_expression)}) has not been normalized, "
                "its resolved type can't be accessed.",
            )
        return self._resolved_type

    def resolve_type(self, plugin: "Plugin", resolver: Namespace) -> inmanta_type.Type:
        """
        Convert the string representation of this argument's type to a type.
        If no type annotation is present or if the type annotation allows any type to be passed
        as argument, then None is returned.

        Expects to be called during the normalization stage.

        :param plugin: The plugin that this argument is part of.
        :param resolver: The namespace that can be used to resolve the type annotation of this
            argument.
        """
        if isinstance(self.type_expression, collections.abc.Hashable) and self.type_expression in PLUGIN_TYPES:
            self._resolved_type = PLUGIN_TYPES[self.type_expression]
            return self._resolved_type

        plugin_line: Range = Range(plugin.location.file, plugin.location.lnr, 1, plugin.location.lnr + 1, 1)
        if not isinstance(self.type_expression, str):
            if typing_inspect.is_union_type(self.type_expression) and not typing.get_args(self.type_expression):
                # If typing.Union is not subscripted, isinstance(self.type_expression, type) evaluates to False.
                raise InvalidTypeAnnotation(stmt=None, msg=f"Union type must be subscripted, got {self.type_expression}")
            if (
                isinstance(self.type_expression, (type, typing.TypeAliasType))
                or typing.get_origin(self.type_expression) is not None
            ):
                self._resolved_type = to_dsl_type(self.type_expression, plugin_line, resolver)
            else:
                raise InvalidTypeAnnotation(
                    stmt=None,
                    msg="Bad annotation in plugin %s for %s, expected str or python type but got %s (%s)"
                    % (plugin.get_full_name(), self.VALUE_NAME, type(self.type_expression).__name__, self.type_expression),
                )
        else:
            locatable_type: LocatableString = LocatableString(self.type_expression, plugin_line, 0, resolver)
            self._resolved_type = inmanta_type.resolve_type(locatable_type, resolver)
        return self._resolved_type

    def validate(self, value: object) -> bool:
        """
        Validate that the given value can be passed to this argument.  Returns True if the value is known
        and valid, False if the value is unknown, and raises a :py:class:`inmanta.ast.RuntimeException`
        if the value is not of the expected type.

        :param value: The value to validate
        """
        if isinstance(value, Unknown):
            # Value is not known, it can not be validated
            return False

        # Validate the value, use custom validate method of the type if it exists
        return self.resolved_type.validate(value)

    @abstractmethod
    def get_signature_display(self, *, use_dsl_types: bool = False) -> str:
        """
        Get the string representing the type for this plugin value.
        The dsl_type argument controls if this type should be returned
        as the plain python type (as it is defined in the plugin e.g. list[str])
        or as the corresponding inferred Inmanta DSL type (e.g. string[]).

        :param use_dsl_types: Display this value's type as the inferred Inmanta DSL
            type or as plain python.
        :return: the string representation for this type in the specified
            language
        """
        raise NotImplementedError()


class PluginArgument(PluginValue):
    """
    Represents the argument of an Inmanta plugin.
    """

    VALUE_TYPE = "argument value"

    # Marker used to indicate that a plugin argument has no default value.
    NO_DEFAULT_VALUE_SET = object()

    def __init__(
        self,
        arg_name: str,
        arg_type: object,
        arg_position: Optional[int] = None,
        default_value: object = NO_DEFAULT_VALUE_SET,
    ) -> None:
        super().__init__(arg_type)
        self.arg_name = arg_name
        self.arg_type = self.type_expression
        self.arg_position = arg_position
        self.is_kw_only_argument = arg_position is None
        self._default_value = default_value
        self.VALUE_NAME = self.arg_name

    @property
    def default_value(self) -> Optional[object]:
        if not self.has_default_value():
            raise Exception("PluginArgument doesn't have a default value")
        return self._default_value

    def has_default_value(self) -> bool:
        """
        Return True iff this plugin argument has a default value set.
        """
        return self._default_value is not self.NO_DEFAULT_VALUE_SET

    def get_signature_display(self, *, use_dsl_types: bool = False) -> str:
        if use_dsl_types:
            return "%s: %s" % (self.arg_name, self.resolved_type.type_string_internal())

        return str(self)

    def __str__(self) -> str:
        if self.has_default_value():
            return "%s: %s = %s" % (self.arg_name, repr(self.arg_type), str(self.default_value))
        else:
            return "%s: %s" % (self.arg_name, repr(self.arg_type))


class PluginReturn(PluginValue):
    """
    Represent the return type of an Inmanta plugin.
    """

    VALUE_TYPE = "returned value"
    VALUE_NAME = "return value"

    def get_signature_display(self, *, use_dsl_types: bool = False) -> str:
        if use_dsl_types:
            return str(self.resolved_type)
        else:
            return repr(self.type_expression)

    def resolve_type(self, plugin: "Plugin", resolver: Namespace) -> inmanta_type.Type:
        out = super().resolve_type(plugin, resolver)
        self._resolved_type = out
        return out


@dataclasses.dataclass
class CheckedArgs:

    args: list[object]
    kwargs: Mapping[str, object]
    unknowns: bool


class Plugin(NamedType, WithComment, metaclass=PluginMeta):
    """
    This class models a plugin that can be called from the language.
    """

    deprecated: bool = False
    replaced_by: Optional[str] = None

    def __init__(self, namespace: Namespace) -> None:
        self.ns = namespace
        self.namespace = namespace

        # The index of the Context attribute.
        self._context: int = -1

        # Load the signature and build all the PluginArgument objects corresponding to it
        self.args: list[PluginArgument] = list()
        self.var_args: Optional[PluginArgument] = None
        self.kwargs: dict[str, PluginArgument] = dict()
        self.var_kwargs: Optional[PluginArgument] = None
        self.all_args: dict[str, PluginArgument] = dict()
        self.return_type: PluginReturn = PluginReturn("null")
        if hasattr(self.__class__, "__function__"):
            self._load_signature(self.__class__.__function__)

        self.new_statement = None

        filename: Optional[str] = inspect.getsourcefile(self.__class__.__function__)
        assert filename is not None
        try:
            line: int = inspect.getsourcelines(self.__class__.__function__)[1] + 1
        except OSError:
            # In case of bytecompiled code there is no source line
            line = 1

        if self.__class__.__function__.__doc__:
            self.comment = self.__class__.__function__.__doc__

        self.location = Location(filename, line)

    def normalize(self) -> None:
        self.resolver = self.namespace

        # Resolve all the types that we expect to receive as input of our plugin
        for arg in self.all_args.values():
            arg.resolve_type(self, self.resolver)
        if self.var_args is not None:
            self.var_args.resolve_type(self, self.resolver)
        if self.var_kwargs is not None:
            self.var_kwargs.resolve_type(self, self.resolver)

        self.return_type.resolve_type(self, self.resolver)

        LOGGER.debug(
            "Found plugin %s::%s",
            self.namespace,
            self.get_signature(use_dsl_types=True),
        )

    def _load_signature(self, function: Callable[..., object]) -> None:
        """
        Load the signature from the given python function, and update the relevant attributes
        of this object.
        """
        # Reset relevant object attributes
        self.args = list()
        self.var_args = None
        self.kwargs = dict()
        self.var_kwargs = None
        self.all_args = dict()

        # Inspect the function to get its arguments and annotations
        arg_spec = inspect.getfullargspec(function)

        # Load the return annotation.  If not return annotation is provided, the returned
        # type is "any".
        if "return" not in arg_spec.annotations:
            self.return_type = PluginReturn("any")
        else:
            self.return_type = PluginReturn(arg_spec.annotations["return"])

        def get_annotation(arg: str) -> object:
            """
            Get the annotation for a specific argument, and if none exists, raise an exception
            """
            if arg not in arg_spec.annotations:
                raise RuntimeException(
                    stmt=None,
                    msg=f"All arguments of plugin {repr(self.get_full_name())} should be annotated: "
                    f"{repr(arg)} has no annotation",
                )

            return arg_spec.annotations[arg]

        if arg_spec.varargs is not None:
            # We have a catch-all positional arguments
            self.var_args = PluginArgument(
                arg_name=arg_spec.varargs,
                arg_type=get_annotation(arg_spec.varargs),
            )

        if arg_spec.varkw is not None:
            # We have a catch-all keyword arguments
            self.var_kwargs = PluginArgument(
                arg_name=arg_spec.varkw,
                arg_type=get_annotation(arg_spec.varkw),
            )

        # Save all positional arguments
        defaults_start_at = len(arg_spec.args) - len(arg_spec.defaults or [])
        for position, arg in enumerate(arg_spec.args):
            annotation = get_annotation(arg)
            if annotation == Context:
                self._context = position
                continue

            # Resolve the default before changing the position because
            # of the context presence, as the defaults list definitely
            # takes into account the context presence.
            default = (
                arg_spec.defaults[position - defaults_start_at]
                if position >= defaults_start_at
                else PluginArgument.NO_DEFAULT_VALUE_SET
            )

            if self._context != -1:
                # If we have a context argument, the position index
                # needs to be adapted as this context object can never be passed
                # from the model.
                position -= 1

            argument = PluginArgument(
                arg_name=arg,
                arg_type=annotation,
                arg_position=position,
                default_value=default,
            )

            # This is a positional argument, we register it now
            self.args.append(argument)
            self.all_args[arg] = argument

        # Save all key-word arguments
        for arg in arg_spec.kwonlyargs:
            argument = PluginArgument(
                arg_name=arg,
                arg_type=get_annotation(arg),
                default_value=(
                    arg_spec.kwonlydefaults[arg]
                    if arg_spec.kwonlydefaults is not None and arg in arg_spec.kwonlydefaults
                    else PluginArgument.NO_DEFAULT_VALUE_SET
                ),
            )
            self.kwargs[arg] = argument
            self.all_args[arg] = argument

    def get_signature(self, *, use_dsl_types: bool = False) -> str:
        """
        Generate the signature of this plugin. Return a string
        containing the arguments and their types, and the type
        of the returned value.  The `dsl_types` argument controls
        whether to display the types in the signature as python
        types, or as the corresponding inferred types from the
        Inmanta DSL.

        :param use_dsl_types: control if the signature should be displayed
        using the plain python types (dsl_types=False) as written in the plugin
        or the corresponding inferred Inmanta DSL types (dsl_types=True).

        """
        # Start the list with all positional arguments
        arg_list = [arg.get_signature_display(use_dsl_types=use_dsl_types) for arg in self.args]

        # Filter all positional arguments out of the kwargs list
        kwargs = [
            arg.get_signature_display(use_dsl_types=use_dsl_types) for _, arg in self.kwargs.items() if arg.is_kw_only_argument
        ]

        if self.var_args is not None:
            arg_list.append("*" + self.var_args.get_signature_display(use_dsl_types=use_dsl_types))
        elif kwargs:
            # For keyword only arguments, we need a marker if we don't have a catch-all
            # positional argument
            arg_list.append("*")

        # Add all keyword-only arguments to the list
        arg_list.extend(kwargs)

        if self.var_kwargs is not None:
            arg_list.append("**" + self.var_kwargs.get_signature_display(use_dsl_types=use_dsl_types))

        # Join all arguments, separated by a comma
        args_string = ", ".join(arg_list)

        if self.return_type.type_expression is None:
            return "%s(%s)" % (self.__class__.__function_name__, args_string)
        return "%s(%s) -> %s" % (
            self.__class__.__function_name__,
            args_string,
            self.return_type.get_signature_display(use_dsl_types=use_dsl_types),
        )

    def get_arg(self, position: int) -> PluginArgument:
        """
        Get the argument at a given position, raise a RuntimeException if it doesn't exists..
        If a catch-all positional argument is defined, it will never raise a RuntimeException.

        :param position: The position of the argument (excluding any Context argument)
        """
        if position < len(self.args):
            return self.args[position]
        elif self.var_args is not None:
            return self.var_args
        else:
            raise RuntimeException(None, f"{self.get_full_name()}() got an unexpected positional argument: {position}")

    def get_kwarg(self, name: str) -> PluginArgument:
        """
        Get the argument with a given name, raise a RuntimeException if it doesn't exists.
        If a catch-all keyword argument is defined, it will never raise a RuntimeException.

        We currently don't support positional-only parameters, so we can simply look for any
        parameter named this way, positional or not.

        :param name: The name of the argument
        """
        if name in self.all_args:
            return self.all_args[name]
        elif self.var_kwargs is not None:
            return self.var_kwargs
        else:
            # Trying to provide a keyword argument which doesn't exist
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(None, f"{self.get_full_name()}() got an unexpected keyword argument: '{name}'")

    def report_missing_arguments(
        self, missing_args: collections.abc.Sequence[str], args_sort: Literal["positional", "keyword-only"]
    ) -> None:
        """
        Helper method to raise an exception specifying that the given arguments are missing.  We try here
        to stick as much as possible to the error that python would have raised, only changing its type.

        The type of the exception raised is RuntimeException.

        If the list of missing_args is empty, we don't report any exception.

        :param missing_args: The missing arguments we should report
        :param args_sort: The sort of argument we are checking (positional or kw)
        """
        func = self.get_full_name()
        if len(missing_args) == 1:
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(None, f"{func}() missing 1 required {args_sort} argument: '{missing_args[0]}'")
        if len(missing_args) > 1:
            arg_names = " and ".join(
                (
                    ", ".join(repr(arg) for arg in missing_args[:-1]),
                    repr(missing_args[-1]),
                )
            )
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(None, f"{func}() missing {len(missing_args)} required {args_sort} arguments: {arg_names}")

    def check_args(self, args: Sequence[object], kwargs: Mapping[str, object]) -> CheckedArgs:
        """
        Check if the arguments of the call match the function signature.

        1. Check if we have too many arguments
        2. Check if we have too few arguments
        3. Check if we have any duplicate arguments (provided as positional and keyword)
        4. Validate the type of each of the provided arguments

        :param args: All the positional arguments to pass on to the plugin function
        :param kwargs: All the keyword arguments to pass on to the plugin function
        """
        if self.var_args is None and len(args) > len(self.args):
            # (1) We got too many positional arguments
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(
                None, f"{self.get_full_name()}() takes {len(self.args)} positional arguments but {len(args)} were given"
            )

        # (2) Check that all positional arguments without a default are provided
        missing_positional_arguments = [
            arg.arg_name
            for position, arg in enumerate(self.args)
            if (
                position >= len(args)  # No input from user in positional args
                and arg.arg_name not in kwargs  # No input from user in keyword args
                and not arg.has_default_value()  # No default value in plugin definition
            )
        ]
        self.report_missing_arguments(missing_positional_arguments, "positional")

        # (2) Check that all keyword arguments without a default are provided
        missing_keyword_arguments = [
            name
            for name, arg in self.kwargs.items()
            if (
                arg.is_kw_only_argument  # The argument was not yet checked as positional arg
                and name not in kwargs  # No input from user in keyword args
                and not arg.has_default_value()  # No default value in plugin definition
            )
        ]
        self.report_missing_arguments(missing_keyword_arguments, "keyword-only")

        converted_args = []
        is_unknown = False

        def reference_exception_msg(value: object, arg: PluginArgument) -> str:
            contains: str = "is" if isinstance(value, Reference) else "contains"
            return (
                f"Value {value!r} for argument {arg.arg_name} of plugin {self.get_full_name()} {contains} a reference."
                " To allow references, use `| Reference[...]` in your type annotation."
            )

        # Validate all positional arguments
        for position, value in enumerate(args):
            # (1) Get the corresponding argument, fails if we don't have one
            arg: PluginArgument = self.get_arg(position)
            result: object
            if isinstance(value, Unknown):
                result = value
                is_unknown = True
            else:
                try:
                    # (4) Validate the input value
                    result = validate_and_convert_to_python_domain(
                        name=arg.arg_name,
                        expected_type=arg.resolved_type,
                        value=value,
                    )
                except (UnsetException, MultiUnsetException):
                    raise
                except UnexpectedReference as e:
                    raise PluginTypeException(
                        stmt=None,
                        msg=reference_exception_msg(value, arg),
                        cause=e,
                    )
                except RuntimeException as e:
                    # some validators do not recognize references specially. Best-effort to raise tailored error message.
                    if isinstance(value, Reference) and not arg.resolved_type.supports_references():
                        raise PluginTypeException(
                            stmt=None,
                            msg=reference_exception_msg(value, arg),
                            cause=e,
                        )
                    raise PluginTypeException(
                        stmt=None,
                        msg=(
                            f"Value {value!r} for argument {arg.arg_name} of plugin "
                            f"{self.get_full_name()} has incompatible type."
                            f" Expected type: {arg.resolved_type.type_string_internal()}"
                        ),
                        cause=e,
                    )
            converted_args.append(result)

        converted_kwargs = {}
        # Validate all kw arguments
        for name, value in kwargs.items():
            # (1) Get the corresponding kwarg, fails if we don't have one
            kwarg: PluginArgument = self.get_kwarg(name)

            # (3) Make sure that our argument is not provided twice
            if kwarg.arg_position is not None and kwarg.arg_position < len(args):
                # The exception raised here tries to match as closely as possible what python
                # would have raised as exception
                raise RuntimeException(None, f"{self.get_full_name()}() got multiple values for argument '{name}'")

            # (4) Validate the input value
            if isinstance(value, Unknown):
                result = value
                is_unknown = True
            else:
                try:
                    result = validate_and_convert_to_python_domain(
                        name=kwarg.arg_name,
                        expected_type=kwarg.resolved_type,
                        value=value,
                    )
                except (UnsetException, MultiUnsetException):
                    raise
                except RuntimeException as e:
                    raise PluginTypeException(
                        stmt=None,
                        msg=(
                            f"Value {value} for argument {kwarg.arg_name} of plugin"
                            f" {self.get_full_name()} has incompatible type."
                            f" Expected type: {kwarg.resolved_type.type_string_internal()}"
                        ),
                        cause=e,
                    )
            converted_kwargs[name] = result

        return CheckedArgs(args=converted_args, kwargs=converted_kwargs, unknowns=is_unknown)

    def emit_statement(self) -> "DynamicStatement":
        """
        This method is called to determine if the plugin call pushes a new
        statement
        """
        return self.new_statement

    def is_accept_unknowns(self) -> bool:
        return self.opts["allow_unknown"]

    def check_requirements(self) -> None:
        """
        Check if the plug-in has all it requires
        """
        if "bin" in self.opts and self.opts["bin"] is not None:
            for _bin in self.opts["bin"]:
                p = subprocess.Popen(["bash", "-c", "type -p %s" % _bin], stdout=subprocess.PIPE)
                result = p.communicate()

                if len(result[0]) == 0:
                    raise Exception(f"{self.__function_name__} requires {_bin} to be available in $PATH")

    @classmethod
    def deprecate_function(cls, replaced_by: Optional[str] = None) -> None:
        cls.deprecated = True
        cls.replaced_by = replaced_by

    def __call__(self, *args: object, **kwargs: object) -> object:
        """
        The function call itself

        As a call, for backward compat. Used by the Jinja template proxy.

        The arguments should already have been passed through `check_args()`
        """
        if self.deprecated:
            msg: str = f"Plugin '{self.get_full_name()}' is deprecated."
            if self.replaced_by:
                msg += f" It should be replaced by '{self.replaced_by}'."
            warnings.warn(PluginDeprecationWarning(msg))
        self.check_requirements()

        def new_arg(arg: object) -> object:
            if isinstance(arg, Context):
                # Not expected to happen, as the compiler itself now uses call_in_context
                return arg
            elif isinstance(arg, Unknown) and self.is_accept_unknowns():
                # If false, DynamicProxy.return_value wil raise an exception
                return arg
            else:
                # call return_value again, just in case. No proxy context needed, because it should really
                # have passed here already
                return DynamicProxy.return_value(arg)

        new_args = [new_arg(arg) for arg in args]
        new_kwargs = {k: new_arg(v) for k, v in kwargs.items()}

        value = self.call(*new_args, **new_kwargs)

        value = DynamicProxy.unwrap(value)

        # Validate the returned value
        try:
            self.return_type.validate(value)
        except (UnsetException, MultiUnsetException):
            raise
        except RuntimeException as e:
            raise PluginTypeException(
                stmt=None,
                msg=(
                    f"Return value {value} of plugin {self.get_full_name()} has incompatible type."
                    f" Expected type: {self.return_type.resolved_type.type_string_internal()}"
                ),
                cause=e,
            )

        return value

    def call_in_context(
        self,
        processed_args: CheckedArgs,
        resolver: Resolver,
        queue: QueueScheduler,
        location: Range,
    ) -> object:
        """
        The function call itself, with compiler context
        """
        if self.deprecated:
            msg: str = f"Plugin '{self.get_full_name()}' is deprecated."
            if self.replaced_by:
                msg += f" It should be replaced by '{self.replaced_by}'."
            warnings.warn(PluginDeprecationWarning(msg))
        self.check_requirements()
        args = processed_args.args
        kwargs = processed_args.kwargs
        value = self.call(*args, **kwargs)

        value = DynamicProxy.unwrap(
            value,
            dynamic_context=proxy.DynamicUnwrapContext(
                resolver=resolver,
                queue=queue,
                location=location,
                type_resolver=functools.partial(to_dsl_type, location=location, resolver=self.namespace),
            ),
        )

        try:
            # Validate the returned value
            self.return_type.validate(value)
            return value
        except (UnsetException, MultiUnsetException):
            raise
        except RuntimeException as e:
            raise PluginTypeException(
                stmt=None,
                msg=(
                    f"Return value {value} of plugin {self.get_full_name()} has incompatible type."
                    f" Expected type: {self.return_type.resolved_type.type_string_internal()}"
                ),
                cause=e,
            )

    def get_full_name(self) -> str:
        return f"{self.ns.get_full_name()}::{self.__class__.__function_name__}"

    def type_string(self) -> str:
        return self.get_full_name()

    def as_python_type_string(self) -> "str | None":
        raise NotImplementedError("Plugins should not be arguments to plugins, this code is not expected to be called")

    def corresponds_to(self, type: inmanta_type.Type) -> bool:
        raise NotImplementedError("Plugins should not be arguments to plugins, this code is not expected to be called")

    def to_python(self, instance: object, *, path: str) -> object:
        raise NotImplementedError("Plugins should not be arguments to plugins, this code is not expected to be called")


@typing.overload
def plugin(
    function: str | None = None,
    commands: Optional[list[str]] = None,
    emits_statements: bool = False,
    allow_unknown: bool = False,
) -> Callable[[T_FUNC], T_FUNC]: ...


@typing.overload
def plugin(
    function: T_FUNC,
    commands: Optional[list[str]] = None,
    emits_statements: bool = False,
    allow_unknown: bool = False,
) -> T_FUNC: ...


@stable_api
def plugin(
    function: T_FUNC | str | None = None,
    commands: Optional[list[str]] = None,
    emits_statements: bool = False,
    allow_unknown: bool = False,
) -> T_FUNC | Callable[[T_FUNC], T_FUNC]:
    """
    Python decorator to register functions with inmanta as plugin

    :param function: The function to register with inmanta. This is the first argument when it is used as decorator.
    :param commands: A list of command paths that need to be available. Inmanta raises an exception when the command is
                     not available.
    :param emits_statements: Set to true if this plugin emits new statements that the compiler should execute. This is only
                             required for complex plugins such as integrating a template engine.
    :param allow_unknown: Set to true if this plugin accepts Unknown values as valid input.
    """

    def curry_name(
        name: Optional[str] = None,
        commands: Optional[list[str]] = None,
        emits_statements: bool = False,
        allow_unknown: bool = False,
    ) -> Callable[[T_FUNC], T_FUNC]:
        """
        Function to curry the name of the function
        """

        def call(fnc: T_FUNC) -> T_FUNC:
            """
            Create class to register the function and return the function itself
            """

            def wrapper(self, *args: object, **kwargs: object) -> Any:
                """
                Python will bind the function as method into the class
                """
                return fnc(*args, **kwargs)

            nonlocal name

            if name is None:
                name = fnc.__name__

            ns_parts = str(fnc.__module__).split(".")
            ns_parts.append(name)
            if ns_parts[0] != const.PLUGINS_PACKAGE:
                raise Exception("All plugin modules should be loaded in the %s package" % const.PLUGINS_PACKAGE)

            fq_plugin_name = "::".join(ns_parts[1:])

            dictionary: dict[str, object] = {}
            dictionary["__module__"] = fnc.__module__

            dictionary["__function_name__"] = name
            dictionary["__fq_plugin_name__"] = fq_plugin_name

            dictionary["opts"] = {"bin": commands, "emits_statements": emits_statements, "allow_unknown": allow_unknown}
            dictionary["call"] = wrapper
            dictionary["__function__"] = fnc

            bases = (Plugin,)
            fnc.__plugin__ = PluginMeta.__new__(PluginMeta, name, bases, dictionary)  # type: ignore[attr-defined]
            return fnc

        return call

    if function is None:
        return curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif isinstance(function, str):
        return curry_name(function, commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif function is not None:
        fnc = curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)
        return fnc(function)


@stable_api
def deprecated(
    function: Optional[Callable] = None, *, replaced_by: Optional[str] = None, **kwargs: abc.Mapping[str, object]
) -> Callable:
    """
    the kwargs are currently ignored but where added in case we want to add something later on.
    """

    def inner(fnc: Callable):
        if hasattr(fnc, "__plugin__"):
            fnc.__plugin__.deprecate_function(replaced_by)
        else:
            raise Exception(
                f"Can not deprecate '{fnc.__name__}': The '@deprecated' decorator should be used in combination with the "
                f"'@plugin' decorator and should be placed at the top."
            )
        return fnc

    if function is not None:
        return inner(function)
    return inner


@stable_api
def allow_reference_values[T](instance: T) -> T:  # T not bound to DynamicProxy because it is not a user-exposed type
    """
    For the given DSL instance, or list or dict nested inside an instance, allow accessing undeclared reference values
    (attributes, list elements or dict values respectively). Reference values are otherwise rejected on access because
    not all plugins can be assumed to be compatible with them, and may reasonably expect values of the DSL attributes' declared
    type.

    Does not allow nested access. Each object for which reference elements are expected should be wrapped separately, e.g.
    `allow_reference_values(my_instance.my_relation).maybe_reference` rather than
    `allow_reference_values(my_instance).my_relation.maybe_reference`.

    This function is not required for plugin arguments or dataclasses. In those cases, reference support can be declared
    directly via type annotations (e.g. `int | Reference[int]`), in which case no special access function is required.
    However, when called on such values, the function simply returns the argument unchanged.
    """
    return instance._allow_references() if isinstance(instance, DynamicProxy) else instance



"""
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """
    Rename some rps columns and values
    """
    schema = """
        ALTER TABLE public.resource_persistent_state RENAME COLUMN blocked_status TO blocked;
        ALTER TABLE public.resource_persistent_state RENAME COLUMN deployment_result TO last_deploy_result;

        UPDATE public.resource_persistent_state
        SET blocked=(
                CASE blocked
                    WHEN 'YES' THEN 'BLOCKED'
                    ELSE 'NOT_BLOCKED'
                END
            )
        ;
    """
    await connection.execute(schema)



"""
Copyright 2022 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncio import Event

"""
This module contains tests related to PostgreSQL and how we interact with it. Its purpose is to verify PostgreSQL internals,
especially those that are not explicitly documented anywhere. These tests act both as a confidence check for compatibility with
future PostgreSQL versions and as a knowledge collector to refer back to when we need it.
"""
import asyncio
from collections import abc

import asyncpg
import pytest


@pytest.mark.slowtest
async def test_postgres_cascade_locking_order(postgresql_pool, run_without_keeping_psql_logs) -> None:
    """
    Verifies that Postgres' cascade deletion acquires locks top-down. This is important because in order to avoid deadlocks
    we define a corresponding locking order for all transactions. See `TableLockMode`, `RowLockMode` and `ConfigurationModel`
    docstrings in `inmanta.data`.
    """
    async with postgresql_pool.acquire() as connection:
        await connection.execute(
            """
            CREATE TABLE IF NOT EXISTS root (name varchar PRIMARY KEY);
            CREATE TABLE IF NOT EXISTS leaf (
                name varchar PRIMARY KEY,
                myroot varchar REFERENCES root(name) ON DELETE CASCADE
            );
            """
        )

    async def insert():
        async with postgresql_pool.acquire() as connection:
            await connection.execute(
                """
                INSERT INTO root VALUES
                    ('root1');

                INSERT INTO leaf VALUES
                    ('leaf1', 'root1'),
                    ('leaf2', 'root1');
                """
            )

    async def lock_top_down():
        async with postgresql_pool.acquire() as connection:
            async with connection.transaction():
                await connection.execute("LOCK TABLE root IN SHARE MODE")
                await asyncio.sleep(0.1)
                await connection.execute("LOCK TABLE leaf IN SHARE MODE")

    async def lock_bottom_up():
        async with postgresql_pool.acquire() as connection:
            async with connection.transaction():
                await connection.execute("LOCK TABLE leaf IN SHARE MODE")
                await asyncio.sleep(0.1)
                await connection.execute("LOCK TABLE root IN SHARE MODE")

    async def delete():
        async with postgresql_pool.acquire() as connection:
            async with connection.transaction():
                await asyncio.sleep(0.05)
                await connection.execute("DELETE FROM root")

    await insert()
    await asyncio.gather(
        lock_top_down(),
        delete(),
    )

    await insert()
    with pytest.raises(asyncpg.DeadlockDetectedError):
        await asyncio.gather(
            lock_bottom_up(),
            delete(),
        )


@pytest.mark.slowtest
@pytest.mark.parametrize("definition_order_one_two", [True, False])
async def test_postgres_cascade_locking_order_siblings(
    postgresql_pool, definition_order_one_two: bool, run_without_keeping_psql_logs
) -> None:
    """
    Verifies locking order for siblings in the cascade tree. Locking order seems to be based on definition order of the
    referencing columns. Locking order may shift in case of updates to the definition of these columns.
    """
    leaf_definitions: tuple[str] = tuple(
        f"CREATE TABLE IF NOT EXISTS {name} (name varchar PRIMARY KEY, myroot varchar REFERENCES root(name) ON DELETE CASCADE);"
        for name in ("leafone", "leaftwo")
    )
    async with postgresql_pool.acquire() as connection:
        await connection.execute(
            """
            CREATE TABLE IF NOT EXISTS root (name varchar PRIMARY KEY);
            %s
            %s
            """
            % (leaf_definitions if definition_order_one_two else tuple(reversed(leaf_definitions)))
        )

    async def insert():
        async with postgresql_pool.acquire() as connection:
            await connection.execute(
                """
                INSERT INTO root VALUES
                    ('root1');

                INSERT INTO leafone VALUES
                    ('leafone1', 'root1'),
                    ('leafone2', 'root1');

                INSERT INTO leaftwo VALUES
                    ('leaftwo1', 'root1'),
                    ('leaftwo2', 'root1');
                """
            )

    async def lock_one_two():
        async with postgresql_pool.acquire() as connection:
            async with connection.transaction():
                await connection.execute("LOCK TABLE leafone IN SHARE MODE")
                await asyncio.sleep(0.1)
                await connection.execute("LOCK TABLE leaftwo IN SHARE MODE")

    async def lock_two_one():
        async with postgresql_pool.acquire() as connection:
            async with connection.transaction():
                await connection.execute("LOCK TABLE leaftwo IN SHARE MODE")
                await asyncio.sleep(0.1)
                await connection.execute("LOCK TABLE leafone IN SHARE MODE")

    async def delete():
        async with postgresql_pool.acquire() as connection:
            async with connection.transaction():
                await asyncio.sleep(0.05)
                await connection.execute("DELETE FROM root")

    success: abc.Awaitable[None] = lock_one_two() if definition_order_one_two else lock_two_one()
    deadlock: abc.Awaitable[None] = lock_two_one() if definition_order_one_two else lock_one_two()

    # always do successful first so we don't need to deal with cleanup after deadlock
    await insert()
    await asyncio.gather(
        success,
        delete(),
    )

    await insert()
    with pytest.raises(asyncpg.DeadlockDetectedError):
        await asyncio.gather(
            deadlock,
            delete(),
        )


@pytest.mark.slowtest
async def test_postgres_transaction_re_entry(postgresql_pool) -> None:
    """
    When do transaction lock each other out?

    More specifically, how to make the release_version method lock itself out.
    """

    # Make a table
    async with postgresql_pool.acquire() as connection:
        await connection.execute("CREATE TABLE IF NOT EXISTS root (name varchar PRIMARY KEY, released BOOL);")
        await connection.execute(
            """
            INSERT INTO root VALUES
                ('root1', False),
                ('root2', False);
            """
        )

    async def update_root(name: str, lock: Event):
        # Main routine: lock table and update if required
        # there is lock given as an argument to make sure we can make one wait for the other
        async with postgresql_pool.acquire() as connection:
            async with connection.transaction():
                print(f"{name}: ENTER")
                # for update is the key here!!!
                record = await connection.fetchrow("select released from root where name=$1 for update", "root1")
                print(f"{name}: WAIT")
                await lock.wait()
                assert len(record) == 1
                if record[0] is True:
                    return False
                print(f"{name}: UPDATE")
                await connection.execute("UPDATE root SET released=true where name=$1", "root1")
                return True
        print(f"{name}: COMMITTED")

    # Two locks
    l1 = Event()
    l2 = Event()

    # Two task running
    f1 = asyncio.create_task(update_root("1", l1))
    f2 = asyncio.create_task(update_root("2", l2))

    # Sleep a bit to get them to lock up
    await asyncio.sleep(0.1)
    # Unlock
    l2.set()
    l1.set()

    # get results
    r1 = await f1
    r2 = await f2

    # One should return true, the other false
    print(r1, r2)
    assert r1 + r2 == 1


@pytest.mark.slowtest
async def test_postgres_index_join_where(postgresql_pool) -> None:
    """
    Verify that the PG analyzer is sufficiently smart to combine constraints from join and where conditions to select
    appropriate indexes, regardless of ordering.
    """

    async with postgresql_pool.acquire() as connection:
        await connection.execute(
            """\
            CREATE TABLE IF NOT EXISTS mytable (x int PRIMARY key, y int, z int);
            CREATE INDEX mytable_x_y_z_index ON mytable (x, y, z);
            INSERT INTO mytable (x, y, z) VALUES (1, 2, 3), (3, 2, 5);
            """
        )
        result = await connection.fetch(
            """\
            EXPLAIN ANALYZE
            SELECT mytable.*
            FROM mytable
            INNER JOIN (VALUES (1, 2, 3)) AS v(x, y, z)
            -- join on x and z, "skipping over" y in the index
            ON mytable.x = v.x AND mytable.z = v.z
            -- where condition on y => index constraints are complete now
            WHERE mytable.y = 2
            """
        )
        assert "Index Only Scan using mytable_x_y_z_index on mytable" in result[0][0]



"""
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import abc
import builtins
import collections
import dataclasses
import hashlib
import json
import typing
import uuid
from typing import Generic, Literal, Never, Optional, Tuple

import pydantic
import typing_inspect
from pydantic import ValidationError

import inmanta
import inmanta.resources
import jsonpath_ng
import typing_extensions
from inmanta import util
from inmanta.types import JsonType, ResourceIdStr, StrictJson

ReferenceType = typing.Annotated[str, pydantic.StringConstraints(pattern="^([a-z0-9_]+::)+[A-Z][A-z0-9_-]*$")]
PrimitiveTypes = str | float | int | bool | None


# The name of an attribute on the class where dataclasses store field information. This is an integral part of the
# dataclasses protocol. All the helpers in dataclasses rely on this field, however `dataclasses._FIELDS` that has
# the same value is not exported.
DATACLASS_FIELDS = "__dataclass_fields__"

# Typing of dataclass.* methods relies entirely on the definition in typeshed that only exists during typechecking.
# This ensures that our code works during typechecking and at runtime.
if not typing.TYPE_CHECKING:
    DataclassProtocol = object
else:
    import _typeshed

    DataclassProtocol = _typeshed.DataclassInstance

    import inmanta.ast.type as inm_type
    from inmanta.agent import handler

type RefValue = PrimitiveTypes | DataclassProtocol


class ReferenceCycleException(Exception):
    """Exception raised when a reference refers to itself"""

    def __init__(self, first_ref: "Reference[RefValue]") -> None:
        self.references: list[Reference[RefValue]] = [first_ref]
        self.complete = False

    def add(self, element: "Reference") -> None:
        """Collect parent entities while traveling up the stack"""
        if self.complete:
            return
        if element in self.references:
            self.complete = True
        self.references.append(element)

    def get_message(self) -> str:
        trace = " -> ".join([repr(x) for x in self.references])
        return "Reference cycle detected: %s" % (trace)

    def __str__(self) -> str:
        return self.get_message()


class Argument(pydantic.BaseModel):
    """Base class for reference (resolver) arguments"""

    type: str
    """ The type of the argument. It is used as a discriminator to select the correct class
    """

    name: str
    """ The name of the argument in the reference constructor
    """

    @abc.abstractmethod
    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        """Get the value for the argument to be able to construct the reference again

        :param resource: The resource on which the reference has been defined
        :param logger: The logger context to use while resolving/deserializing
        """


class LiteralArgument(Argument):
    """Literal argument to a reference"""

    type: typing.Literal["literal"] = "literal"
    value: PrimitiveTypes

    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        return self.value


class JsonArgument(Argument):
    """Json-like argument to a reference"""

    type: typing.Literal["json"] = "json"
    value: StrictJson

    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        return self.value


class ReferenceArgument(Argument):
    """Use the value of another reference as an argument for the reference"""

    type: typing.Literal["reference"] = "reference"
    id: uuid.UUID

    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        return resource.get_reference_value(self.id, logger)


class MutatedJsonArgument(Argument):
    """
    Json-like argument that contains another reference in the json structure

    It is stored as
    1. a json value with all reference replaced by None
    2. a mapping of dictpaths to references. Each dict path indicates where the reference should be inserted

    """

    type: typing.Literal["mjson"] = "mjson"
    value: StrictJson
    references: dict[str, ReferenceArgument]

    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        start_value: JsonType = self.value
        for destination, valueref in self.references.items():
            value = valueref.get_arg_value(resource, logger)

            # backward compat between 8.2 and higher
            # jsonpath_ng does not allow starting with ., dictpath does
            if destination.startswith("."):
                destination = "$" + destination
            jsonpath_expr = jsonpath_ng.parse(destination)
            jsonpath_expr.update(start_value, value)
        return start_value


class GetArgument(Argument):
    """Get a value from the resource body as value"""

    type: typing.Literal["get"] = "get"
    dict_path_expression: str

    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        return None


class PythonTypeArgument(Argument):
    """Use the python type as an argument. This is mostly used for references that are generic for a
    type, such as core::AttributeReference
    """

    type: typing.Literal["python_type"] = "python_type"
    value: str

    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        return getattr(builtins, self.value)


class ResourceArgument(Argument):
    """This argument provides the resource itself to the reference"""

    type: typing.Literal["resource"] = "resource"
    id: ResourceIdStr

    def get_arg_value(
        self,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> object:
        if not resource.id.resource_str() == self.id:
            raise Exception(
                f"This resource refers to another resource {self.id} instead of "
                f"itself {resource.id.resource_str()}, this is not supported"
            )
        return resource


ArgumentTypes = typing.Annotated[
    LiteralArgument
    | ReferenceArgument
    | GetArgument
    | PythonTypeArgument
    | ResourceArgument
    | JsonArgument
    | MutatedJsonArgument,
    pydantic.Field(discriminator="type"),
]
""" A list of all specific types of arguments. Pydantic uses this to instantiate the correct argument class
"""


class SerializedReferenceLike(pydantic.BaseModel):
    """
    A base model class for references and mutators in their serialized form
    """

    type: ReferenceType
    args: list[ArgumentTypes]


class ReferenceModel(SerializedReferenceLike):
    """A reference"""

    id: uuid.UUID


class MutatorModel(SerializedReferenceLike):
    """A mutator"""


C = typing.TypeVar("C", bound="ReferenceLike")


CYCLE_TOKEN = object()
# Token to perform cycle detection when serializing


class ReferenceLike:
    """A base class for references and mutators"""

    type: typing.ClassVar[ReferenceType]

    def __init__(self) -> None:
        self._model: typing.Optional[SerializedReferenceLike] | Literal[CYCLE_TOKEN] = None

        # Only present in compiler
        # Will be set by DynamicProxy.unwrap at the plugin boundary
        self._model_type: typing.Optional["inm_type.Type"] = None

    def resolve_other[S: RefValue](self, value: "Reference[S] | S", logger: "handler.LoggerABC") -> S:
        """
        Given a reference or a value, either return the value or the value obtained by resolving the reference.

        This method is intended to be used by mutators and references to resolve their own parameters
        (which could be references as well.

        When resolving references on the handler side, all parameters will already be resolved
        and this method only serves to ensure correct typing.

        When resolving reference on the compiler side, not all references will be resolved and using this method is required.
        """
        if isinstance(value, Reference):
            return value.get(logger)
        return value

    @classmethod
    def deserialize(
        cls: typing.Type[C],
        ref: SerializedReferenceLike,
        resource: "inmanta.resources.Resource",
        logger: "handler.LoggerABC",
    ) -> C:
        """Deserialize the reference or mutator.

        :param ref: The model of the reference to deserialize
        :param resource: The resource to use as a read-only reference
        """
        return cls(**{arg.name: arg.get_arg_value(resource, logger) for arg in ref.args})

    @abc.abstractmethod
    def serialize(self) -> SerializedReferenceLike:
        """Serialize to be able to add them to a resource"""
        raise NotImplementedError()

    def serialize_arguments(self) -> Tuple[uuid.UUID, list[ArgumentTypes]]:
        """Serialize the arguments to this class"""

        # The handling of references here is a bit subtle:
        # We replace every reference with a ReferenceArgument that refers to the reference by id
        # The reference itself is not handled here but in inmanta.resources.ReferenceSubCollector.collect_reference
        # There, the raw, unserialized tree of arguments is iterated over as well to collect the references themselves
        # The caches on the Reference prevent this from being too inefficient by serializing only once
        arguments: list[ArgumentTypes] = []
        for name, value in self.arguments.items():
            match value:
                case str() | int() | float() | bool() | None:
                    arguments.append(LiteralArgument(name=name, value=value))
                case dict() | list():
                    collector = inmanta.resources.ReferenceSubCollector()
                    # The collector here is purely to collect the path/reference pairs
                    # The set of reference it collects will be discarded
                    # The root ReferenceCollector will traverse past this point as well to collect the actual reference
                    cleaned_value = collector.collect_references(value, "$")
                    try:
                        if collector.references:
                            arguments.append(
                                MutatedJsonArgument(
                                    name=name,
                                    value=cleaned_value,
                                    references={
                                        path: ReferenceArgument(name=path, id=model.id)
                                        for path, model in collector.replacements.items()
                                    },
                                )
                            )
                        else:
                            arguments.append(JsonArgument(name=name, value=value))
                    except ValidationError:
                        raise ValueError(f"The {name} attribute of {self!r} is not json serializable: {value}")
                case Reference():
                    model = value.serialize()
                    arguments.append(ReferenceArgument(name=name, id=model.id))

                case inmanta.resources.Resource() as v:
                    arguments.append(ResourceArgument(name=name, id=v.id.resource_str()))
                case type() if value in [str, float, int, bool]:
                    arguments.append(PythonTypeArgument(name=name, value=value.__name__))

                case _:
                    raise TypeError(f"Unable to serialize argument `{name}` of `{self!r}` with value {value}")

        data = json.dumps({"type": self.type, "args": arguments}, default=util.api_boundary_json_encoder, sort_keys=True)
        hasher = hashlib.md5()
        hasher.update(data.encode())
        return uuid.uuid3(uuid.NAMESPACE_OID, hasher.digest()), arguments

    @property
    def arguments(self) -> collections.abc.Mapping[str, object]:
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other):
            return False

        assert isinstance(other, ReferenceLike)  # mypy can't figure out the check above

        return self.arguments == other.arguments


class Mutator(ReferenceLike):
    """A mutator that has side effects when executed"""

    @abc.abstractmethod
    def run(self, logger: "handler.LoggerABC") -> None:
        """Execute the mutator"""

    def serialize(self) -> MutatorModel:
        """Emit the correct pydantic objects to serialize the reference in the exporter."""
        if not self._model:
            arg_id, arguments = self.serialize_arguments()
            self._model = MutatorModel(type=self.type, args=arguments)

        assert isinstance(self._model, MutatorModel)
        return self._model


T = typing_extensions.TypeVar("T", bound=RefValue, covariant=True, default=RefValue)


class Reference(ReferenceLike, Generic[T]):
    """Instances of this class can create references to a value and resolve them."""

    def __init__(self) -> None:
        super().__init__()
        self._reference_value: T
        self._reference_value_cached: bool = False

    @classmethod
    def get_reference_type(cls) -> type[T] | None:
        # We are that inherits from Reference[T]
        # We do a best effort here to untangle this, but it is difficult because of
        # https://github.com/ilevkivskyi/typing_inspect/issues/110
        # We can only handle the case where T is a concrete type, not where it is a re-mapped type-var
        # https://github.com/inmanta/inmanta-core/issues/8765
        for g in cls.__orig_bases__:  # type: ignore
            if typing_inspect.is_generic_type(g) and typing.get_origin(g) is Reference:
                return typing.get_args(g)[0]
        return None

    @classmethod
    def is_dataclass_reference(cls) -> bool:
        return dataclasses.is_dataclass(cls.get_reference_type())

    @abc.abstractmethod
    def resolve(self, logger: "handler.LoggerABC") -> T:
        """This method resolves the reference and returns the object that it refers to"""
        pass

    def get(self, logger: "handler.LoggerABC") -> T:
        """Get the value. If we have already resolved it a cached value is returned, otherwise resolve() is called"""
        if not self._reference_value_cached:
            self._reference_value = self.resolve(logger)
            self._reference_value_cached = True

        else:
            logger.debug("Using cached value for reference %(reference)s", reference=repr(self))
        return self._reference_value

    def serialize(self) -> ReferenceModel:
        """Emit the correct pydantic objects to serialize the reference in the exporter."""
        if self._model is CYCLE_TOKEN:
            raise ReferenceCycleException(self)
        if not self._model:
            self._model = CYCLE_TOKEN
            try:
                argument_id, arguments = self.serialize_arguments()
            except ReferenceCycleException as e:
                e.add(self)
                raise
            self._model = ReferenceModel(id=argument_id, type=self.type, args=arguments)

        assert isinstance(self._model, ReferenceModel)
        return self._model

    def __bool__(self) -> Never:
        raise NotImplementedError(f"{self!r} is an inmanta reference, not a boolean.")


class reference:
    """This decorator registers a reference under a specific name"""

    # It is not allowed to use T in a class var so we cannot use T here
    _reference_classes: typing.ClassVar[dict[str, type[Reference[RefValue]]]] = {}

    def __init__(self, name: str) -> None:
        """
        :param name: This name is used to indicate the type of the reference
        """
        self.name = name

    @classmethod
    def add_reference(cls, name: str, reference_type: type[Reference[RefValue]]) -> None:
        cls._reference_classes[name] = reference_type

    def __call__[C: type[Reference]](self, cls: C) -> C:
        """Register a new reference. If we already have it explicitly delete it (reload)"""
        if self.name in type(self)._reference_classes:
            del type(self)._reference_classes[self.name]

        type(self)._reference_classes[self.name] = cls
        cls.type = self.name
        return cls

    @classmethod
    def get_class(cls, name: str) -> type[Reference[RefValue]]:
        """Get the reference class registered with the given name"""
        if name not in cls._reference_classes:
            raise TypeError(f"There is no reference class registered with name {name}")

        return cls._reference_classes[name]

    @classmethod
    def reset(cls) -> None:
        """Reset the registered reference classes"""
        # Keep core ones
        cls._reference_classes = {k: v for k, v in cls._reference_classes.items() if k.startswith("core::")}

    @classmethod
    def get_references(cls) -> typing.Iterator[tuple[str, type[Reference[RefValue]]]]:
        """Return an iterator with all items registered."""
        return (item for item in cls._reference_classes.items())


class mutator:
    """This decorator register a mutator under a specific name"""

    _mutator_classes: typing.ClassVar[dict[str, type[Mutator]]] = {}

    def __init__(self, name: str) -> None:
        """
        :param name: This name is used to indicate the type of the mutator
        """
        self.name = name

    @classmethod
    def add_mutator(cls, name: str, mutator_type: type[Mutator]) -> None:
        cls._mutator_classes[name] = mutator_type

    def __call__[T: Mutator](self, cls: type[T]) -> type[T]:
        """Register a new mutator. If we already have it explicitly delete it (reload)"""
        if self.name in mutator._mutator_classes:
            del mutator._mutator_classes[self.name]

        mutator._mutator_classes[self.name] = cls
        cls.type = self.name
        return cls

    @classmethod
    def get_class(cls, name: str) -> type[Mutator]:
        """Get the mutator class registered with the given name"""
        if name not in cls._mutator_classes:
            raise TypeError(f"There is no mutator class registered with name {name}")

        return cls._mutator_classes[name]

    @classmethod
    def reset(cls) -> None:
        """Reset the registered mutator classes"""
        # Keep core ones
        cls._mutator_classes = {k: v for k, v in cls._mutator_classes.items() if k.startswith("core::")}

    @classmethod
    def get_mutators(cls) -> typing.Iterator[tuple[str, type[Mutator]]]:
        """Return an iterator with all items registered."""
        return (item for item in cls._mutator_classes.items())


@reference(name="core::AttributeReference")
class AttributeReference[T: PrimitiveTypes](Reference[T]):
    """A reference that points to a value in an attribute of a dataclass"""

    def __init__(
        self,
        reference: DataclassProtocol | Reference[DataclassProtocol],
        attribute_name: str,
    ) -> None:
        super().__init__()
        self.attribute_name = attribute_name
        self.reference = reference

    def resolve(self, logger: "handler.LoggerABC") -> T:
        return typing.cast(T, getattr(self.resolve_other(self.reference, logger), self.attribute_name))


@mutator(name="core::Replace")
class ReplaceValue(Mutator):
    """Replace a reference in the provided resource"""

    def __init__(self, resource: "inmanta.resources.Resource", value: Reference[PrimitiveTypes], destination: str) -> None:
        """Change a value in the given resource at the given distination

        :param resource: The resource to replace the value in
        :param value: The value to replace in `resource` at `destination`
        :param destination: A dictpath expression where to replace the value
        """
        super().__init__()
        self.resource = resource
        self.value = value
        self.destination = destination

    def run(self, logger: "handler.LoggerABC") -> None:
        value = self.resolve_other(self.value, logger)
        jsonpath_expr = jsonpath_ng.parse(self.destination)
        jsonpath_expr.update(self.resource, value)


@typing.runtime_checkable
class MaybeReference(typing.Protocol):
    """
    DSL value that may represent a reference in the Python domain, while having a different value in the DSL domain.

    This includes DSL dataclass instances with reference attributes, if they were initially constructed in the Python
    domain as a reference to a dataclass instance (and converted on the boundary).
    """

    __slots__ = ()

    def unwrap_reference(self) -> Optional[Reference]:
        """
        If this DSL value represents a reference value, returns the associated reference object. Otherwise returns None.
        """
        ...


def unwrap_reference(value: object) -> Optional[Reference]:
    """
    Iff the given value is a reference or a DSL value that represents a reference, returns the associated reference.
    Otherwise returns None.

    This includes DSL dataclass instances with reference attributes, if they were initially constructed in the Python
    domain as a reference to a dataclass instance (and converted on the boundary).
    """
    return value if isinstance(value, Reference) else value.unwrap_reference() if isinstance(value, MaybeReference) else None


def is_reference_of(instance: typing.Optional[object], type_class: type[object]) -> bool:
    """Is the given instance a reference to the given type."""
    if instance is None or not isinstance(instance, Reference):
        return False

    return instance.get_reference_type() == type_class



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

import time
from threading import Lock

from .. import Clock
from ..stats.moving_average import ExpWeightedMovingAvg


class Meter(object):
    """
    A meter metric which measures mean throughput and one-, five-, and fifteen-minute
    exponentially-weighted moving average throughputs.
    """

    def __init__(self, clock: Clock = time) -> None:
        super(Meter, self).__init__()
        self.lock = Lock()
        self.clock = clock
        self.clear()

    def clear(self) -> None:
        with self.lock:
            self.start_time = self.clock.time()
            self.counter = 0.0
            self.m1rate = ExpWeightedMovingAvg(period=1, clock=self.clock)
            self.m5rate = ExpWeightedMovingAvg(period=5, clock=self.clock)
            self.m15rate = ExpWeightedMovingAvg(period=15, clock=self.clock)

    def get_one_minute_rate(self) -> float:
        return self.m1rate.get_rate()

    def get_five_minute_rate(self) -> float:
        return self.m5rate.get_rate()

    def get_fifteen_minute_rate(self) -> float:
        return self.m15rate.get_rate()

    def tick(self) -> None:
        self.m1rate.tick()
        self.m5rate.tick()
        self.m15rate.tick()

    def mark(self, value: float = 1) -> None:
        with self.lock:
            self.counter += value
            self.m1rate.add(value)
            self.m5rate.add(value)
            self.m15rate.add(value)

    def get_count(self) -> float:
        return self.counter

    def get_mean_rate(self) -> float:
        if self.counter == 0:
            return 0
        elapsed: float = self.clock.time() - self.start_time
        return self.counter / elapsed



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import logging
import sys
import uuid
from typing import Collection

import inmanta.loader
from inmanta import protocol
from inmanta.agent import executor
from inmanta.agent.executor import ResourceInstallSpec
from inmanta.data.model import LEGACY_PIP_DEFAULT, PipConfig
from inmanta.protocol import Client, SyncClient
from inmanta.types import ResourceType
from inmanta.util.async_lru import async_lru_cache

LOGGER = logging.getLogger(__name__)


class CouldNotResolveCode(Exception):

    def __init__(self, resource_type: str, version: int, error_message: str) -> None:
        self.msg = f"Failed to get source code for {resource_type} version={version}, result={error_message}"
        super().__init__(self.msg)


class CodeManager:
    """
    Helper responsible for translating resource versions into code

    Caches heavily
    """

    def __init__(self, client: Client) -> None:
        self._client = client

    @async_lru_cache(maxsize=5)
    async def get_pip_config(self, environment: uuid.UUID, version: int) -> PipConfig:
        response = await self._client.get_pip_config(tid=environment, version=version)
        if response.code != 200:
            raise Exception("Could not get pip config from server " + str(response.result))
        assert response.result is not None  # mypy
        pip_config = response.result["data"]
        if pip_config is None:
            return LEGACY_PIP_DEFAULT
        return PipConfig(**pip_config)

    @async_lru_cache(maxsize=1024)
    async def get_code_for_type(self, environment: uuid.UUID, version: int, resource_type: ResourceType) -> ResourceInstallSpec:
        result: protocol.Result = await self._client.get_source_code(environment, version, resource_type)
        if result.code == 200 and result.result is not None:
            sync_client = SyncClient(client=self._client, ioloop=asyncio.get_running_loop())
            requirements: set[str] = set()
            sources: list["inmanta.loader.ModuleSource"] = []
            # Encapsulate source code details in ``ModuleSource`` objects
            module_sources = result.result["data"]["sources"]
            project_constraints = result.result["data"]["project_constraints"]
            for source in module_sources:
                sources.append(
                    inmanta.loader.ModuleSource(
                        name=source["module_name"],
                        is_byte_code=source["is_byte_code"],
                        hash_value=source["hash"],
                        _client=sync_client,
                    )
                )
                requirements.update(source["requirements"])

            resource_install_spec = ResourceInstallSpec(
                resource_type,
                version,
                executor.ExecutorBlueprint(
                    environment_id=environment,
                    pip_config=await self.get_pip_config(environment, version),
                    requirements=list(requirements),
                    sources=sources,
                    python_version=sys.version_info[:2],
                    project_constraints=project_constraints if project_constraints else None,
                ),
            )
            return resource_install_spec
        else:
            raise CouldNotResolveCode(resource_type, version, str(result.get_result()))

    async def get_code(
        self, environment: uuid.UUID, version: int, resource_types: Collection[ResourceType]
    ) -> tuple[Collection[ResourceInstallSpec], executor.FailedResources]:
        """
        Get the collection of installation specifications (i.e. pip config, python package dependencies,
        Inmanta modules sources) required to deploy a given version for the provided resource types.

        Expects at least one resource type.

        :return: Tuple of:
            - collection of ResourceInstallSpec for resource_types with valid handler code and pip config
            - set of invalid resource_types (no handler code and/or invalid pip config)
        """
        if not resource_types:
            raise ValueError(f"{self.__class__.__name__}.get_code() expects at least one resource type")

        resource_install_specs: list[ResourceInstallSpec] = []
        invalid_resources: executor.FailedResources = {}
        for resource_type in set(resource_types):
            try:
                resource_install_specs.append(await self.get_code_for_type(environment, version, resource_type))
            except CouldNotResolveCode as e:
                LOGGER.error(
                    "%s",
                    e.msg,
                )
                invalid_resources[resource_type] = e
            except Exception as e:
                LOGGER.error(
                    "Failed to get source code for %s version=%d",
                    resource_type,
                    version,
                    exc_info=True,
                )
                invalid_resources[resource_type] = e

        return resource_install_specs, invalid_resources



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import logging

import pytest

from inmanta.agent import reporting
from inmanta.agent.handler import HandlerContext, InvalidOperation
from inmanta.data.model import AttributeStateChange
from inmanta.resources import Id, PurgeableResource
from inmanta.server import SLICE_SESSION_MANAGER

logger = logging.getLogger(__name__)


@pytest.mark.slowtest
async def test_agent_get_status(server, environment, agent):
    clients = server.get_slice(SLICE_SESSION_MANAGER)._sessions.values()
    assert len(clients) == 1
    clients = [x for x in clients]
    client = clients[0].get_client()
    status = await client.get_status()
    status = status.get_result()
    for name in reporting.reports.keys():
        assert name in status and status[name] != "ERROR"
    assert status.get("env") is None


def test_context_changes():
    """Test registering changes in the handler context"""
    resource = PurgeableResource(Id.parse_id("std::testing::NullResource[agent,name=test],v=1"))
    ctx = HandlerContext(resource)

    # use attribute change attributes
    ctx.update_changes({"value": AttributeStateChange(current="a", desired="b")})
    assert len(ctx.changes) == 1

    # use dict
    ctx.update_changes({"value": dict(current="a", desired="b")})
    assert len(ctx.changes) == 1
    assert isinstance(ctx.changes["value"], AttributeStateChange)

    # use dict with empty string
    ctx.update_changes({"value": dict(current="", desired="value")})
    assert len(ctx.changes) == 1
    assert ctx.changes["value"].current == ""
    assert ctx.changes["value"].desired == "value"

    # use tuple
    ctx.update_changes({"value": ("a", "b")})
    assert len(ctx.changes) == 1
    assert isinstance(ctx.changes["value"], AttributeStateChange)

    # use wrong arguments
    with pytest.raises(InvalidOperation):
        ctx.update_changes({"value": ("a", "b", 3)})

    with pytest.raises(InvalidOperation):
        ctx.update_changes({"value": ["a", "b"]})

    with pytest.raises(InvalidOperation):
        ctx.update_changes({"value": "test"})


def test_entrypoint_for_handler_typing():
    from inmanta.agent.agent import AgentInstance  # noqa: F401
